from plantcv.learn.naive_bayes import naive_bayes
from plantcv.learn.naive_bayes import naive_bayes_multiclass
from plantcv.learn.train_kmeans import train_kmeans
from plantcv.learn.train_kmeans import patch_extract

__all__ = ["naive_bayes", "naive_bayes_multiclass", "train_kmeans", "patch_extract"]



from plantcv.plantcv.io.read_dataset import read_dataset
from plantcv.plantcv.io.random_subset import random_subset
from plantcv.plantcv.io.open_url import open_url

# add new functions to end of lists
__all__ = ["read_dataset", "random_subset", "open_url"]



"""Find tips from skeleton image."""
import os
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _find_tips


def find_tips(skel_img, mask=None, label=None):
    """Find tips in skeletonized image.
    The endpoints algorithm was inspired by Jean-Patrick Pommier: https://gist.github.com/jeanpat/5712699

    Inputs:
    skel_img    = Skeletonized image
    mask        = (Optional) binary mask for debugging. If provided, debug image will be overlaid on the mask.
    label       = Optional label parameter, modifies the variable name of
                  observations recorded (default = pcv.params.sample_label).
    Returns:
    tip_img   = Image with just tips, rest 0

    :param skel_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param label: str
    :return tip_img: numpy.ndarray
    """
    tip_img, tip_list, tip_labels = _find_tips(skel_img=skel_img, mask=mask)

    _debug(visual=tip_img, filename=os.path.join(params.debug_outdir, f"{params.device}_skeleton_tips.png"))

    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # Save coordinates to Outputs
    outputs.add_observation(sample=label, variable='tips', trait='list of tip coordinates identified from a skeleton',
                            method='plantcv.plantcv.morphology.find_tips', scale='pixels', datatype=list,
                            value=tip_list, label=tip_labels)

    return tip_img



# Dilation filter

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _dilate
from plantcv.plantcv import params


def dilate(gray_img, ksize, i, roi=None):
    """
    Performs morphological 'dilation' filtering. Adds pixel to center of kernel if conditions set in kernel are true.

    Inputs:
    gray_img = Grayscale (usually binary) image data
    ksize   = Kernel size (int). A k x k kernel will be built. Must be greater than 1 to have an effect.
    i        = iterations, i.e. number of consecutive filtering passes
    roi      = rectangular ROI to dilate within

    Returns:
    dil_img = dilated image

    :param gray_img: numpy.ndarray
    :param ksize: int
    :param i: int
    :param roi: plantcv.plantcv.Objects
    :return dil_img: numpy.ndarray
    """
    dil_img = _dilate(gray_img, ksize, i, roi=roi)

    _debug(visual=dil_img,
           filename=os.path.join(params.debug_outdir,
                                 str(params.device) + '_dil_image' + str(ksize) + '_itr' + str(i) + '.png'),
           cmap='gray')

    return dil_img



# PlantCV classes
import os
import cv2
import json
import numpy as np
import datetime
from plantcv.plantcv import __version__ as ver
from plantcv.plantcv import fatal_error
from plantcv.plantcv.annotate.points import _find_closest_pt
import matplotlib.pyplot as plt
from math import floor
import altair as alt
import pandas as pd


class Params:
    """PlantCV parameters class."""

    def __init__(self, device=0, debug=None, debug_outdir=".", line_thickness=5,
                 line_color=(255, 0, 255), dpi=100, text_size=0.55,
                 text_thickness=2, marker_size=60, color_scale="gist_rainbow", color_sequence="sequential",
                 sample_label="default", saved_color_scale=None, verbose=True, unit="pixels", px_height=1, px_width=1):
        """Initialize parameters.

        Keyword arguments/parameters:
        device            = Device number. Used to count steps in the pipeline. (default: 0)
        debug             = None, print, or plot. Print = save to file, Plot = print to screen. (default: None)
        debug_outdir      = Debug images output directory. (default: .)
        line_thickness    = Width of line drawings. (default: 5)
        line_color        = Color of line annotations (default = (255, 0, 255))
        dpi               = Figure plotting resolution, dots per inch. (default: 100)
        text_size         = Size of plotting text. (default: 0.55)
        text_thickness    = Thickness of plotting text. (default: 2)
        marker_size       = Size of plotting markers (default: 60)
        color_scale       = Name of plotting color scale (matplotlib colormap). (default: gist_rainbow)
        color_sequence    = Build color scales in "sequential" or "random" order. (default: sequential)
        sample_label      = Sample name prefix. Used in analyze functions. (default: "default")
        saved_color_scale = Saved color scale that will be applied next time color_palette is called. (default: None)
        verbose           = Whether or not in verbose mode. (default: True)
        unit              = Units of size trait outputs. (default: "pixels")
        px_height         = Size scaling information about pixel height (default: 1)
        px_width          = Size scaling information about pixel width (default: 1)


        :param device: int
        :param debug: str
        :param debug_outdir: str
        :param line_thickness: numeric
        :param dpi: int
        :param text_size: float
        :param text_thickness: int
        :param marker_size: int
        :param color_scale: str
        :param color_sequence: str
        :param sample_label: str
        :param saved_color_scale: list
        :param verbose: bool
        :param unit: str
        :param px_height: float
        :param px_width: float

        """
        self.device = device
        self.debug = debug
        self.debug_outdir = debug_outdir
        self.line_thickness = line_thickness
        self.line_color = line_color
        self.dpi = dpi
        self.text_size = text_size
        self.text_thickness = text_thickness
        self.marker_size = marker_size
        self.color_scale = color_scale
        self.color_sequence = color_sequence
        self.sample_label = sample_label
        self.saved_color_scale = saved_color_scale
        self.verbose = verbose
        self.unit = unit
        self.px_height = px_height
        self.px_width = px_width


class Outputs:
    """PlantCV outputs class"""

    def __init__(self):
        self.measurements = {}
        self.images = []
        self.observations = {}
        self.metadata = {}

        # Add a method to clear measurements
    def clear(self):
        """Clear all measurements"""
        self.measurements = {}
        self.images = []
        self.observations = {}
        self.metadata = {}

    # Method to add observation to outputs
    def add_observation(self, sample, variable, trait, method, scale, datatype, value, label):
        """Keyword arguments/parameters:
        sample       = Sample name. Used to distinguish between multiple samples
        variable     = A local unique identifier of a variable, e.g. a short name,
                       that is a key linking the definitions of variables with observations.
        trait        = A name of the trait mapped to an external ontology; if there is no exact mapping, an informative
                       description of the trait.
        method       = A name of the measurement method mapped to an external ontology; if there is no exact mapping, an
                       informative description of the measurement procedure
        scale        = Units of the measurement or scale in which the observations are expressed; if possible, standard
                       units and scales should be used and mapped to existing ontologies; in the case of non-standard
                       scale a full explanation should be given
        datatype     = The type of data to be stored, e.g. 'int', 'float', 'str', 'list', 'bool', etc.
        value        = The data itself
        label        = The label for each value (most useful when the data is a frequency table as in hue,
                       or other tables)

        :param sample: str
        :param variable: str
        :param trait: str
        :param method: str
        :param scale: str
        :param datatype: type
        :param value:
        :param label:
        """
        # Create an empty dictionary for the sample if it does not exist
        if sample not in self.observations:
            self.observations[sample] = {}

        # Validate that the data type is supported by JSON
        _ = _validate_data_type(value)

        # Save the observation for the sample and variable
        self.observations[sample][variable] = {
            "trait": trait,
            "method": method,
            "scale": scale,
            "datatype": str(datatype),
            "value": value,
            "label": label
        }

    # Method to add metadata instance to outputs
    def add_metadata(self, term, datatype, value):
        """Add a metadata term and value to outputs.

        Parameters
        ----------
        term : str
            Metadata term/name.
        datatype : type
            The type of data to be stored, e.g. 'int', 'float', 'str', 'list', 'bool', etc.
        value : any
            The data itself.
        """
        # Create an empty dictionary for the sample if it does not exist
        if term not in self.metadata:
            self.metadata[term] = {}

        # Validate that the data type is supported by JSON
        _ = _validate_data_type(value)

        # Save the observation for the sample and variable
        self.metadata[term] = {
            "datatype": str(datatype),
            "value": [value]
        }

    # Method to save observations to a file
    def save_results(self, filename, outformat="json"):
        """Save results to a file.

        Keyword arguments/parameters:
        filename       = Output filename
        outformat      = Output file format ("json" or "csv"). Default = "json"

        :param filename: str
        :param outformat: str
        """
        # Add current date & time to metadata in UTC format
        run_datetime = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        self.add_metadata(term="run_date", datatype=str, value=run_datetime)
        self.add_metadata(term="plantcv_version", datatype=str, value=ver)

        if outformat.upper() == "JSON":
            if os.path.isfile(filename):
                with open(filename, 'r') as f:
                    hierarchical_data = json.load(f)
                    hierarchical_data["observations"] = self.observations
                    existing_metadata = hierarchical_data["metadata"]
                    for term in self.metadata:
                        save_term = term
                        if term in existing_metadata:
                            save_term = f"{term}_1"
                        hierarchical_data["metadata"][save_term] = self.metadata[term]
            else:
                hierarchical_data = {"metadata": self.metadata, "observations": self.observations}
            with open(filename, mode='w') as f:
                json.dump(hierarchical_data, f)

        elif outformat.upper() == "CSV":
            # Open output CSV file
            with open(filename, "w") as csv_table:
                # Gather any additional metadata
                metadata_key_list = list(self.metadata.keys())
                metadata_val_list = [val["value"] for val in self.metadata.values()]
                # Write the header
                header = metadata_key_list + ["sample", "trait", "value", "label"]
                csv_table.write(",".join(map(str, header)) + "\n")
                # Iterate over data samples
                for sample in self.observations:
                    # Iterate over traits for each sample
                    for var in self.observations[sample]:
                        val = self.observations[sample][var]["value"]
                        # If the data type is a list or tuple we need to unpack the data
                        if isinstance(val, (list, tuple)):
                            # Combine each value with its label
                            for value, label in zip(self.observations[sample][var]["value"],
                                                    self.observations[sample][var]["label"]):
                                # Skip list of tuple data types
                                if not isinstance(value, tuple):
                                    # Save one row per value-label
                                    row = metadata_val_list + [sample, var, value, label]
                                    csv_table.write(",".join(map(str, row)) + "\n")
                        # If the data type is Boolean, store as a numeric 1/0 instead of True/False
                        elif isinstance(val, bool):
                            row = metadata_val_list + [sample, var, int(self.observations[sample][var]["value"]),
                                                       self.observations[sample][var]["label"]]
                            csv_table.write(",".join(map(str, row)) + "\n")
                        # For all other supported data types, save one row per trait
                        # Assumes no unusual data types are present (possibly a bad assumption)
                        else:
                            row = metadata_val_list + [sample, var, self.observations[sample][var]["value"],
                                                       self.observations[sample][var]["label"]]
                            csv_table.write(",".join(map(str, row)) + "\n")

    def plot_dists(self, variable):
        """Plot a distribution of data.

        Keyword arguments/parameters:
        variable      = A local unique identifier of a variable, e.g. a short name,
                        that is a key linking the definitions of variables with observations.
        Returns:
        chart          = Altair chart object
        :param variable: str
        :return chart: altair.vegalite.v4.api.Chart
        """
        alt.data_transformers.disable_max_rows()
        data = {"sample": [], "value": [], "label": []}
        # Iterate over measurement sample groups
        for sample in self.observations:
            # If the measurement variable is present in the sample
            # And the data type is a list
            if variable in self.observations[sample] and "list" in (self.observations[sample][variable]["datatype"]):
                data["value"] = data["value"] + self.observations[sample][variable]["value"]
                data["label"] = data["label"] + self.observations[sample][variable]["label"]
                data["sample"] = data["sample"] + [sample] * len(self.observations[sample][variable]["value"])
        df = pd.DataFrame(data)
        step = 10
        overlap = 10
        chart = alt.Chart(df, height=step, width=500).mark_area(
            interpolate="monotone", fillOpacity=0.8, stroke='lightgray', strokeWidth=0.5
        ).encode(
            alt.X('label:Q').title('Bin labels'),
            alt.Y('value:Q').axis(None).scale(range=[step, -step * overlap])
        ).facet(
            alt.Row('sample:N').title(None).header(labelAngle=0, labelAlign='left', labelOrient="left")
        ).configure_facet(
            spacing=0,
            columns=1
        ).properties(
            bounds='flush'
        ).configure_title(
            anchor="end"
        ).configure_view(
            stroke=None
        ).configure_axis(
            grid=False
        )
        return chart


def _validate_data_type(data):
    """Validate that the data type is supported by JSON.

    Parameters
    ----------
    data : any
        Data to be validated.

    Returns
    -------
    bool
        True if the data type is supported by JSON.

    Raises
    ------
    ValueError
        If the data type is not supported by JSON.
    """
    # Supported data types
    supported_dtype = ["int", "float", "str", "list", "bool", "tuple", "dict", "NoneType", "numpy.float64"]
    # Supported class types
    class_list = [f"<class '{cls}'>" for cls in supported_dtype]

    # Send an error message if datatype is not supported by json
    if str(type(data)) not in class_list:
        # String list of supported types
        type_list = ', '.join(map(str, supported_dtype))
        fatal_error(f"The Data type {type(data)} is not compatible with JSON! Please use only these: {type_list}!")

    return True


class Spectral_data:
    """PlantCV Hyperspectral data class"""

    def __init__(self, array_data, max_wavelength, min_wavelength, max_value, min_value, d_type, wavelength_dict,
                 samples, lines, interleave, wavelength_units, array_type, pseudo_rgb, filename, default_bands,
                 metadata=None):
        # The actual array/datacube
        self.array_data = array_data
        # Min/max available wavelengths (for spectral datacube)
        self.max_wavelength = max_wavelength
        self.min_wavelength = min_wavelength
        #  Min/max pixel value for single wavelength or index
        self.max_value = max_value
        self.min_value = min_value
        # Numpy data type
        self.d_type = d_type
        # Contains all available wavelengths where keys are wavelength and value are indices
        self.wavelength_dict = wavelength_dict
        # Resolution of a single band of spectral data is (samples, lines) rather than (x,y) with other arrays
        self.samples = samples
        self.lines = lines
        # Interleave type
        self.interleave = interleave
        self.wavelength_units = wavelength_units
        # The type of array data (entire datacube, specific index, first derivative, etc)
        self.array_type = array_type
        # Pseudo-RGB image if the array_type is a datacube
        self.pseudo_rgb = pseudo_rgb
        # The filename where the data originated from
        self.filename = filename
        # The default band indices needed to make an pseudo_rgb image, if not available then store None
        self.default_bands = default_bands
        # Metadata, flexible components in a dictionary
        self.metadata = metadata
        if not metadata:
            self.metadata = {}


class PSII_data:
    """PSII data class"""

    def __init__(self):
        self.ojip_dark = None
        self.ojip_light = None
        self.pam_dark = None
        self.pam_light = None
        self.spectral = None
        self.chlorophyll = None
        self.datapath = None
        self.filename = None

    def __repr__(self):
        mvars = []
        for k, v in self.__dict__.items():
            if v is not None:
                mvars.append(k)
        return "PSII variables defined:\n" + '\n'.join(mvars)

    def add_data(self, protocol):
        """Input:
        protocol: xr.DataArray with name equivalent to initialized attributes
        """
        self.__dict__[protocol.name] = protocol


class Points:
    """Point annotation/collection class to use in Jupyter notebooks. It allows the user to
    interactively click to collect coordinates from an image. Left click collects the point and
    right click removes the closest collected point
    """

    def __init__(self, img, figsize=(12, 6)):
        """
        Initialization
        :param img: image data
        :param figsize: desired figure size, (12,6) by default
        :attribute points: list of points as (x,y) coordinates tuples
        """
        self.fig, self.ax = plt.subplots(1, 1, figsize=figsize)
        self.ax.imshow(cv2.cvtColor(img, cv2.COLOR_BGR2RGB))

        self.points = []
        self.events = []

        self.fig.canvas.mpl_connect('button_press_event', self.onclick)

    def onclick(self, event):
        """Handle mouse click events"""
        self.events.append(event)
        if event.button == 1:

            self.ax.plot(event.xdata, event.ydata, 'x', c='red')
            self.points.append((floor(event.xdata), floor(event.ydata)))

        else:
            idx_remove, _ = _find_closest_pt((event.xdata, event.ydata), self.points)
            # remove the closest point to the user right clicked one
            self.points.pop(idx_remove)
            self.ax.lines[idx_remove].remove()
        self.fig.canvas.draw()


class Objects:
    """Class for managing image contours/objects and their hierarchical relationships."""

    def __init__(self, contours: list = None, hierarchy: list = None):
        self.contours = contours
        self.hierarchy = hierarchy
        self._n = 0
        if contours is None:
            self.contours = []
            self.hierarchy = []

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        if self._n < len(self.contours):
            self._n += 1
            return Objects(contours=[self.contours[self._n-1]], hierarchy=[self.hierarchy[self._n-1]])
        raise StopIteration

    def append(self, contour, h):
        """Append a contour and its hierarchy to the object"""
        self.contours.append(contour)
        self.hierarchy.append(h)

    def save(self, filename):
        """Save the object to a file"""
        np.savez(filename, contours=self.contours, hierarchy=self.hierarchy)

    @staticmethod
    def load(filename):
        """Load a saved object file"""
        file = np.load(filename)
        obj = Objects(file['contours'].tolist(), file['hierarchy'])
        return obj



"""Find tangent angles in degrees of skeleton segments."""
import os
import cv2
import numpy as np
import pandas as pd
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import color_palette
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _iterative_prune


def _slope_to_intersect_angle(m1, m2):
    """
    Calculate the intersection angle (in degrees) between two lines given their slopes.

    Parameters
    ----------
    m1 : float
        Slope of the first line.
    m2 : float
        Slope of the second line.

    Returns
    -------
    angle : float
        Intersection angle between the two lines, in degrees.
    """
    angle = ((np.pi - np.absolute(np.arctan(m1) - np.arctan(m2))) * 180 / np.pi).astype(np.float64)
    return angle


def segment_tangent_angle(segmented_img, objects, size, label=None):
    """
    Find 'tangent' angles in degrees of skeleton segments.

    Uses `size` pixels on either end of each segment to find a linear regression line, and calculates the
    angle between the two lines drawn per segment.

    Parameters
    ----------
    segmented_img : numpy.ndarray
        Segmented image to plot slope lines and intersection angles on.
    objects : list
        List of contours.
    size : int
        Size of ends used to calculate "tangent" lines.
    label : str, optional
        Label parameter, modifies the variable name of observations recorded (default = pcv.params.sample_label).

    Returns
    -------
    labeled_img : numpy.ndarray
        Segmented debugging image with angles labeled.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    # Store debug
    debug = params.debug
    params.debug = None

    labeled_img = segmented_img.copy()
    intersection_angles = []
    label_coord_x = []
    label_coord_y = []

    # Create a color scale, use a previously stored scale if available
    rand_color = color_palette(num=len(objects), saved=True)

    for i, cnt in enumerate(objects):
        find_tangents = np.zeros(segmented_img.shape[:2], np.uint8)
        cv2.drawContours(find_tangents, objects, i, 255, 1, lineType=8)
        cv2.drawContours(labeled_img, objects, i, rand_color[i], params.line_thickness, lineType=8)
        pruned_segment = _iterative_prune(find_tangents, size)
        segment_ends = find_tangents - pruned_segment
        segment_end_obj, _ = _cv2_findcontours(bin_img=segment_ends)
        slopes = []
        for obj in segment_end_obj:
            # Find bounds for regression lines to get drawn
            rect = cv2.minAreaRect(cnt)
            pts = cv2.boxPoints(rect)
            df = pd.DataFrame(pts, columns=('x', 'y'))
            x_max = int(df['x'].max())
            x_min = int(df['x'].min())

            # Find line fit to each segment
            vx, vy, x, y = cv2.fitLine(obj, cv2.DIST_L2, 0, 0.01, 0.01)
            slope = -vy / vx
            left_list = int(np.array(((x - x_min) * slope) + y).item())
            right_list = int(np.array(((x - x_max) * slope) + y).item())
            slopes.append(slope)

            if abs(slope) > 1000000:
                print("Slope of contour with ID#", i, "is", slope, "and cannot be plotted.")
            else:
                # Draw slope lines
                cv2.line(labeled_img, (x_max - 1, right_list), (x_min, left_list), rand_color[i], 1)

        if len(slopes) < 2:
            # If size*2>len(obj) then pruning will remove the segment completely, and
            # makes segment_end_objs contain just one contour.
            print("Size too large, contour with ID#", i, "got pruned away completely.")
            intersection_angles.append("NA")
        else:
            # Calculate intersection angles
            slope1 = slopes[0][0]
            slope2 = slopes[1][0]
            intersection_angle = _slope_to_intersect_angle(slope1, slope2)
            intersection_angles.append(intersection_angle)

        # Store coordinates for labels
        label_coord_x.append(objects[i][0][0][0])
        label_coord_y.append(objects[i][0][0][1])

    segment_ids = []
    # Reset debug mode
    params.debug = debug

    for i, cnt in enumerate(objects):
        # Label slope lines
        w = label_coord_x[i]
        h = label_coord_y[i]
        if type(intersection_angles[i]) is str:
            text = f"{intersection_angles[i]}"
        else:
            text = f"{intersection_angles[i]:0,.2f}"
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(i)

    outputs.add_observation(sample=label, variable='segment_tangent_angle', trait='segment tangent angle',
                            method='plantcv.plantcv.morphology.segment_tangent_angle', scale='degrees', datatype=list,
                            value=intersection_angles, label=segment_ids)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segment_tangent_angles.png"))

    return labeled_img



"""Convert RGB to CMYK colorspace and return the specified subchannel as a grayscale image."""

import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.plantcv._helpers import _rgb2cmyk


def rgb2gray_cmyk(rgb_img, channel):
    """Convert image from RGB colorspace to CMYK colorspace. Returns the specified subchannel as a gray image.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data
    channel : str
        color subchannel (c = cyan, m = magenta, y = yellow, k=black)

    Returns
    -------
    numpy.ndarray
        grayscale image from one CMYK color channel
    """
    # Convert RGB to CMYK and return the specified subchannel as a gray image
    gray_img = _rgb2cmyk(rgb_img=rgb_img, channel=channel)

    # The allowable channel inputs are c, m , y or k
    names = {"c": "cyan", "m": "magenta", "y": "yellow", "k": "black"}

    # Save or display the grayscale image
    _debug(visual=gray_img,
           filename=os.path.join(params.debug_outdir, f"{params.device}_cmyk_{names[channel.lower()]}.png"), cmap="gray")

    return gray_img



# Filter objects based on calculated properties
import os
import numpy as np
from skimage.measure import label, regionprops
from plantcv.plantcv import params, fatal_error
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace


def obj_props(bin_img, cut_side="upper", thresh=0, regprop="area", roi=None):
    """Detect/filter regions in a binary image based on calculated properties.

    Parameters:
    ----------
    bin_img : numpy.ndarray
        Binary image containing the objects to consider.
    cut_side : str, default: "upper"
        Side to keep when objects are divided by the "thresh" value.
    thresh : int | float, default: 0
        Region property threshold value.
    regprop : str, default: "area"
        Region property to filter on. Can choose from "area" or other int and float properties calculated by
        skimage.measure.regionprops.
    roi : plantcv.plantcv.Objects, default None
        Optional region of interest to apply the object property filter within

    Returns:
    -------
    filtered_mask : numpy.ndarray
        Binary image that contains only the filtered objects.
    """
    # Increment step counter
    params.device += 1
    # Make cut_side all lowercase
    cut_side = cut_side.lower()
    # Check if cut_side is valid
    if cut_side not in ("upper", "lower"):
        fatal_error("Must specify either 'upper' or 'lower' for cut_side")
    # subset binary image for ROI
    sub_bin_img = _rect_filter(bin_img, roi=roi)
    # Skip empty masks
    if np.count_nonzero(sub_bin_img) != 0:
        # label connected regions in ROI
        labeled_img = label(sub_bin_img)
        # measure region properties
        obj_measures = regionprops(labeled_img)
        # list of correct data types
        correct_types = [np.int64, np.float64, int, float]
        # check to see if property of interest is the right type
        if type(getattr(obj_measures[0], regprop)) not in correct_types:
            fatal_error(f"Property {regprop} is not an integer or float type.")

        # blank mask to draw discs onto
        sub_filtered_mask = np.zeros(labeled_img.shape, dtype=np.uint8)
        # Pull all values and calculate the mean
        valueslist = []
        # Store the list of coordinates (row,col) for the objects that pass
        for obj in obj_measures:
            # Object color
            gray_val = 255
            # Store the value of the property for each object
            valueslist.append(getattr(obj, regprop))
            # If it is an upper threshold, keep the objects that are above the threshold
            if cut_side == "upper":
                gray_val = 255 if getattr(obj, regprop) > thresh else 0
            # If it is a lower threshold, keep the objects that are below the threshold
            elif cut_side == "lower":
                gray_val = 255 if getattr(obj, regprop) < thresh else 0
            # Add the object to the filtered mask (255 if it passes, 0 if it does not)
            sub_filtered_mask += np.where(labeled_img == obj.label, gray_val, 0).astype(np.uint8)

        if params.debug == "plot":
            print(f"Min value = {min(valueslist)}")
            print(f"Max value = {max(valueslist)}")
            print(f"Mean value = {sum(valueslist)/len(valueslist)}")

    else:
        sub_filtered_mask = np.copy(sub_bin_img)
    # slice subset back into full size binary image
    filtered_mask = _rect_replace(bin_img, sub_filtered_mask, roi)

    _debug(visual=filtered_mask, filename=os.path.join(params.debug_outdir,
                                                       f"{params.device}_filter_mask_{regprop}_{thresh}.png"))
    return filtered_mask



"""Analyze stem characteristics."""
import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _scale_size


def analyze_stem(rgb_img, stem_objects, label=None):
    """
    Analyze stem characteristics by fitting a linear regression line to stem segments.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image to plot debug image.
    stem_objects : list
        List of stem segments (output from segment_sort function).
    label : str, optional
        Label parameter, modifies the variable name of observations recorded.

    Returns
    -------
    labeled_img : numpy.ndarray
        Stem analysis debugging image.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    labeled_img = np.copy(rgb_img)
    img_x = np.shape(labeled_img)[0]
    grouped_stem = np.vstack(stem_objects)

    # Find vertical height of the stem by measuring bounding box
    stem_x, stem_y, _, height = cv2.boundingRect(grouped_stem)

    # Calculate stem angle
    [vx, vy, x, y] = cv2.fitLine(grouped_stem, cv2.DIST_L2, 0, 0.01, 0.01)
    slope = -vy / vx

    # Calculate stem path length
    stem_length = cv2.arcLength(grouped_stem, False) / 2

    outputs.add_observation(sample=label, variable='stem_height', trait='vertical length of stem segments',
                            method='plantcv.plantcv.morphology.analyze_stem', scale=params.unit, datatype=float,
                            value=_scale_size(height), label=params.unit)
    outputs.add_observation(sample=label, variable='stem_angle', trait='angle of combined stem object',
                            method='plantcv.plantcv.morphology.analyze_stem', scale='degrees', datatype=float,
                            value=float(slope.item()), label='degrees')
    outputs.add_observation(sample=label, variable='stem_length', trait='path length of combined stem object',
                            method='plantcv.plantcv.morphology.analyze_stem', scale=params.unit, datatype=float,
                            value=_scale_size(stem_length), label=params.unit)

    # Draw culm_height
    cv2.line(labeled_img, (int(stem_x), stem_y), (int(stem_x), stem_y + height), (0, 255, 0), params.line_thickness)
    # Draw combined stem angle
    x_min = 0  # Set bounds for regression lines to get drawn
    x_max = img_x
    intercept1 = int(np.array(((x - x_min) * slope) + y).item())
    intercept2 = int(np.array(((x - x_max) * slope) + y).item())
    if abs(slope) > 1000000:
        print("Slope  is ", slope, " and cannot be plotted.")
    else:
        cv2.line(labeled_img, (x_max - 1, intercept2), (x_min, intercept1), (0, 0, 255), 1)
    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_stem_analze.png"))

    return labeled_img



# Debugging module
from plantcv.plantcv import params
from plantcv.plantcv import print_image
from plantcv.plantcv import plot_image


def _debug(visual, filename=None, **kwargs):
    """
    Save or display a visual for debugging.

    Inputs:
    visual   - An image or plot to display for debugging
    filename - An optional filename to save the visual to (default: None)
    kwargs - key-value arguments to xarray.plot method

    :param visual: numpy.ndarray
    :param filename: str
    :param kwargs: dict
    """
    # Auto-increment the device counter
    params.device += 1

    if params.debug == "print":
        # If debug is print, save the image to a file
        print_image(img=visual, filename=filename, **kwargs)
    elif params.debug == "plot":
        # If debug is plot, print to the plotting device
        plot_image(img=visual, **kwargs)



# Find NIR image

import os
from plantcv.plantcv import print_image
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def output_mask(img, mask, filename, outdir=None, mask_only=False):
    """Prints ori image and mask to directories.

    Inputs:
    img = original image, read in with plantcv function read_image
    mask  = binary mask image (single chanel)
    filename = vis image file name (output of plantcv read_image function)
    outdir = output directory
    mask_only = bool for printing only mask

    Returns:
    imgpath = path to image
    maskpath path to mask

    :param img: numpy.ndarray
    :param mask: numpy.ndarray
    :param filename: str
    :param outdir: str
    :param mask_only: bool
    :return imgpath: str
    :return maskpath: str
    """
    analysis_images = []

    directory = os.getcwd() if outdir is None else outdir

    # Return values
    results = []

    # Save the original image unless mask_only=True
    if not mask_only:
        path = os.path.join(str(directory), "ori-images")
        os.makedirs(path, exist_ok=True)
        imgpath = os.path.join(str(path), str(filename))
        print_image(img, imgpath)
        analysis_images.append(['IMAGE', 'ori-img', imgpath])
        results.append(imgpath)
        # Print/plot original image
        _debug(visual=img, filename=os.path.join(params.debug_outdir, f"{params.device}_ori-img.png"))

    # Save the mask
    path = os.path.join(str(directory), "mask-images")
    os.makedirs(path, exist_ok=True)
    maskpath = os.path.join(str(path), str(filename))
    print_image(mask, maskpath)
    analysis_images.append(['IMAGE', 'mask', maskpath])
    results.append(maskpath)
    # Print/plot mask
    _debug(visual=mask, filename=os.path.join(params.debug_outdir, f"{params.device}_mask-img.png"))

    results.append(analysis_images)

    return results



"""Crop an image."""
import os
import cv2
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params
from plantcv.plantcv import Spectral_data


def crop(img, x, y, h, w):
    """Crop image.

    Inputs:
    img       = RGB, grayscale, or hyperspectral image data
    x         = X coordinate of starting point
    y         = Y coordinate of starting point
    h         = Height
    w         = Width

    Returns:
    cropped   = cropped image

    :param img: numpy.ndarray
    :param x: int
    :param y: int
    :param h: int
    :param w: int
    :return cropped: numpy.ndarray
    """
    # Check if the array data format
    if isinstance(img, Spectral_data):
        # Copy the HSI array
        array = np.copy(img.array_data)
        # Copy the pseudo_rgb image
        ref_img = np.copy(img.pseudo_rgb)
        # Crop the HSI data
        cropped_array = array[y:y + h, x:x + w, :]
        # Calculate the array dimensions
        dims1 = cropped_array.shape

        # Create a new Spectral_data object
        cropped = Spectral_data(
            array_data=cropped_array,
            min_wavelength=img.min_wavelength,
            max_wavelength=img.max_wavelength,
            min_value=np.min(cropped_array),
            max_value=np.max(cropped_array),
            d_type=cropped_array.dtype,
            wavelength_dict=img.wavelength_dict,
            samples=int(dims1[1]),
            lines=int(dims1[0]),
            interleave=img.interleave,
            wavelength_units=img.wavelength_units,
            array_type=img.array_type,
            pseudo_rgb=ref_img[y:y + h, x:x + w, :],
            default_bands=img.default_bands,
            filename=img.filename
        )

    else:
        if len(np.shape(img)) > 2 and np.shape(img)[-1] > 3:
            ref_img = img[:, :, [0]]
            ref_img = np.transpose(np.transpose(ref_img)[0])
            cropped = img[y:y + h, x:x + w, :]
        else:
            ref_img = np.copy(img)
            cropped = img[y:y + h, x:x + w]

    # Create the rectangle contour vertices
    pt1 = (x, y)
    pt2 = (x + w - 1, y + h - 1)

    ref_img = cv2.rectangle(img=np.copy(ref_img), pt1=pt1, pt2=pt2, color=(255, 0, 0), thickness=params.line_thickness)

    _debug(visual=ref_img, filename=os.path.join(params.debug_outdir, str(params.device) + "_crop.png"))

    return cropped



"""Find curvature measure of skeleton segments."""
import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import outputs
from plantcv.plantcv import color_palette
from plantcv.plantcv.morphology import segment_path_length
from plantcv.plantcv.morphology import segment_euclidean_length
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _cv2_findcontours, _find_tips


def segment_curvature(segmented_img, objects, label=None):
    """Calculate segment curvature as defined by the ratio between geodesic and euclidean distance.
    Measurement of two-dimensional tortuosity.

    Inputs:
    segmented_img     = Segmented image to plot lengths on
    objects           = List of contours
    label             = Optional label parameter, modifies the variable name of
                        observations recorded (default = pcv.params.sample_label).

    Returns:
    labeled_img        = Segmented debugging image with curvature labeled

    :param segmented_img: numpy.ndarray
    :param objects: list
    :param label: str
    :return labeled_img: numpy.ndarray
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    label_coord_x = []
    label_coord_y = []
    labeled_img = segmented_img.copy()

    # Store debug
    debug = params.debug
    params.debug = None

    _ = segment_euclidean_length(segmented_img, objects, label="backend")
    _ = segment_path_length(segmented_img, objects, label="backend")
    eu_lengths = outputs.observations['backend']['segment_eu_length']['value']
    path_lengths = outputs.observations['backend']['segment_path_length']['value']
    curvature_measure = [float(x / y) for x, y in zip(path_lengths, eu_lengths)]
    # Create a color scale, use a previously stored scale if available
    rand_color = color_palette(num=len(objects), saved=True)

    for i, obj in enumerate(objects):
        # Store coordinates for labels
        label_coord_x.append(obj[0][0][0])
        label_coord_y.append(obj[0][0][1])

        # Draw segments one by one to group segment tips together
        finding_tips_img = np.zeros(segmented_img.shape[:2], np.uint8)
        cv2.drawContours(finding_tips_img, objects, i, (255, 255, 255), 1, lineType=8)
        segment_tips, _, _ = _find_tips(finding_tips_img)
        tip_objects, _ = _cv2_findcontours(bin_img=segment_tips)
        points = []

        for t in tip_objects:
            # Gather pairs of coordinates
            x, y = t.ravel()
            coord = (x, y)
            points.append(coord)

        # Draw euclidean distance lines
        cv2.line(labeled_img, points[0], points[1], rand_color[i], 1)

    segment_ids = []
    # Reset debug mode
    params.debug = debug

    for i, _ in enumerate(objects):
        # Calculate geodesic distance
        text = f"{curvature_measure[i]:0,.3f}"
        w = label_coord_x[i]
        h = label_coord_y[i]
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=(150, 150, 150), thickness=params.text_thickness)
        segment_ids.append(i)

    outputs.add_observation(sample=label, variable='segment_curvature', trait='segment curvature',
                            method='plantcv.plantcv.morphology.segment_curvature', scale='none', datatype=list,
                            value=curvature_measure, label=segment_ids)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_segment_curvature.png"))

    return labeled_img



from plantcv.plantcv.transform.color_correction import get_color_matrix
from plantcv.plantcv.transform.color_correction import get_matrix_m
from plantcv.plantcv.transform.color_correction import calc_transformation_matrix
from plantcv.plantcv.transform.color_correction import apply_transformation_matrix
from plantcv.plantcv.transform.color_correction import save_matrix
from plantcv.plantcv.transform.color_correction import load_matrix
from plantcv.plantcv.transform.color_correction import correct_color
from plantcv.plantcv.transform.color_correction import create_color_card_mask
from plantcv.plantcv.transform.color_correction import quick_color_check
from plantcv.plantcv.transform.find_color_card import find_color_card
from plantcv.plantcv.transform.color_correction import std_color_matrix
from plantcv.plantcv.transform.color_correction import affine_color_correction
from plantcv.plantcv.transform.detect_color_card import detect_color_card
from plantcv.plantcv.transform.rescale import rescale
from plantcv.plantcv.transform.rotate import rotate
from plantcv.plantcv.transform.nonuniform_illumination import nonuniform_illumination
from plantcv.plantcv.transform.resize import resize, resize_factor
from plantcv.plantcv.transform.warp import warp, warp_align
from plantcv.plantcv.transform.gamma_correct import gamma_correct
from plantcv.plantcv.transform.checkerboard_calib import checkerboard_calib, calibrate_camera
from plantcv.plantcv.transform.merge_images import merge_images
from plantcv.plantcv.transform.auto_correct_color import auto_correct_color
from plantcv.plantcv.transform.detect_color_card import mask_color_card

__all__ = ["get_color_matrix", "get_matrix_m", "calc_transformation_matrix", "apply_transformation_matrix",
           "save_matrix", "load_matrix", "correct_color", "create_color_card_mask", "quick_color_check",
           "find_color_card", "std_color_matrix", "affine_color_correction", "rescale", "nonuniform_illumination", "resize",
           "resize_factor", "warp", "rotate", "warp", "warp_align", "gamma_correct", "detect_color_card", "checkerboard_calib",
           "calibrate_camera", "merge_images", "auto_correct_color", "mask_color_card"]



import os
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _closing


def closing(gray_img, kernel=None, roi=None):
    """Closes holes, removing small dark spots (i.e. pepper).

    Inputs:
    gray_img = input image (grayscale or binary)
    kernel   = optional neighborhood, expressed as an array of 1s and 0s. If None, use cross-shaped structuring element.
    roi      = optional rectangular ROI to apply closing within a region

    :param gray_img: np.ndarray
    :param kernel = np.ndarray
    :param roi: plantcv.plantcv.Objects
    :return filtered_img: np.ndarray
    """
    filtered_img = _closing(gray_img, kernel, roi=roi)

    _debug(visual=filtered_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_closing' + '.png'),
           cmap='gray')

    return filtered_img



"""Converter functions."""
import os
import json
import itertools
import pandas as pd
import numpy as np


def json2csv(json_file, csv_prefix):
    """Convert a PlantCV JSON file to a CSV files.

    Parameters
    ----------
    json_file : str
        JSON file output by plantcv-run-workflow.
    csv_prefix : str
        CSV output files prefix.

    Raises
    ------
    IOError
        JSON file does not exist.
    ValueError
        JSON file is not in the expected format.
    """
    if not os.path.exists(json_file):
        # If the file does not exist raise an error
        raise IOError(f"File does not exist: {json_file}")
    # Open the JSON file for reading
    with open(json_file, "r") as fp:
        data = json.load(fp)
    # If the data is JSON but it does not have the components we expect from PlantCV raise an error
    if "variables" not in data or "entities" not in data:
        raise ValueError(f"Invalid JSON file: {json_file}")

    # Split up variables
    meta_vars, scalar_vars, multi_vars = _unpack_variables(data)

    # Output files
    scalar_file = csv_prefix + "-single-value-traits.csv"
    multi_file = csv_prefix + "-multi-value-traits.csv"

    # Create a CSV file of vector traits
    with open(multi_file, "w") as csv:
        # Create a header for the long-format table
        csv.write(",".join(map(str, meta_vars + ["sample", "trait", "value", "label"])) + "\n")
        # Iterate over each entity
        for entity in data["entities"]:
            # Add metadata variables
            meta_row = _create_metadata_row(meta_vars=meta_vars, metadata=entity["metadata"])
            # Add trait variables
            for sample, var in itertools.product(entity["observations"].keys(), multi_vars):
                data_rows = _create_data_rows(var=var, obs=entity["observations"][sample])
                for row in data_rows:
                    csv.write(",".join(map(str, meta_row + [sample] + row)) + "\n")

    # Create a CSV file of scalar traits
    # Initialize a dictionary to store the data
    scalar_data = {
        "sample": [],
        "trait": [],
        "value": [],
        "label": []
    }
    # Add metadata terms to the dictionary
    for var in meta_vars:
        scalar_data[var] = []
    # Iterate over each entity
    for entity in data["entities"]:
        # Add metadata variables
        meta_row = _create_metadata_row(meta_vars=meta_vars, metadata=entity["metadata"])
        # Add trait variables
        for sample, var in itertools.product(entity["observations"].keys(), scalar_vars):
            data_rows = _create_data_rows(var=var, obs=entity["observations"][sample])
            for row in data_rows:
                scalar_data["sample"].append(sample)
                scalar_data["trait"].append(row[0])
                scalar_data["value"].append(row[1])
                scalar_data["label"].append(row[2])
                # Add metadata variables
                for i, _ in enumerate(meta_vars):
                    scalar_data[meta_vars[i]].append(meta_row[i])
    # Create a pandas dataframe from the dictionary
    df = pd.DataFrame(scalar_data)
    # Pivot the dataframe to wide format
    # If duplicate indices exist, use the last set of observations are kept
    df = df.pivot_table(index=meta_vars + ["sample"], columns=["trait", "label"], values="value", aggfunc=_last_index)
    # Collapse hierarchical column headers
    df.columns = [f"{c[0]}_{c[1]}" for c in df.columns]
    # Save the dataframe to a CSV file
    df.to_csv(scalar_file)


def _unpack_variables(data):
    """Unpack variables from "variables" key of outputs
    Parameters
    ----------
    data       = dict, json file

    Return
    ------
    meta_vars   = list, variable names that are metadata
    scalar_vars = list, variable names that are single-value traits
    multi_vars  = list, variable names that are multi-value traits

    """
    # Split up variables
    meta_vars = []
    scalar_vars = []
    multi_vars = []
    for key, var in data["variables"].items():
        # Metadata variables
        if var["category"] == "metadata":
            meta_vars.append(key)
        # Data variables
        else:
            # Scalar variables
            if var["datatype"] in ["<class 'bool'>", "<class 'int'>", "<class 'float'>", "<class 'str'>",
                                   "<type 'bool'>", "<type 'int'>", "<type 'float'>", "<type 'str'>"]:
                scalar_vars.append(key)
            # Vector variables
            if var["datatype"] in ["<class 'list'>", "<type 'list'>", "<class 'tuple'>", "<type 'tuple'>"]:
                multi_vars.append(key)
    return meta_vars, scalar_vars, multi_vars


def _last_index(*args):
    """Aggregation function to return the last index of a list."""
    return np.array(args[-1])[-1]


def _create_metadata_row(meta_vars, metadata):
    """Create a row of metadata.

    Parameters
    ----------
    meta_vars : list
        List of metadata terms
    metadata : dict
        Metadata dictionary

    Returns
    -------
    list
        List of metadata values
    """
    meta_row = []
    for var in meta_vars:
        val = "NA"
        if var in metadata:
            vals = metadata[var]["value"]
            vals = ["none" if v is None else v for v in vals]
            # Create a unique list if there are multiple values
            u_list = np.unique(vals).tolist()
            # If there are multiple values, join them with an underscore
            val = "_".join(map(str, u_list))
        meta_row.append(val)
    return meta_row


def _create_data_rows(var, obs):
    """Create rows of data for a variable.

    Parameters
    ----------
    var : str
        Variable name
    obs : dict
        Data dictionary

    Returns
    -------
    list
        List of lists of data rows
    """
    data_rows = []
    if var in obs:
        value = obs[var]["value"]
        label = obs[var]["label"]
        if isinstance(value, bool):
            value = int(value)
        if isinstance(value, (list, tuple)):
            for val, lbl in zip(value, label):
                if not isinstance(val, tuple):
                    data_rows.append([var, val, lbl])
        else:
            data_rows.append([var, value, label])
    else:
        data_rows.append([var, "NA", "NA"])
    return data_rows


def tabulate_bayes_classes(input_file, output_file):
    """Tabulate pixel RGB values into a table for naive Bayes training.
    Inputs:
    input_file   = Input text file of class names and RGB values
    output_file  = Output file for storing the tab-delimited naive Bayes training data

    The input file should have class names preceded by the "#" character. RGB values can be pasted
    directly from ImageJ without reformatting. E.g.:

    #plant
    96,154,72	95,153,72	91,155,71	91,160,70	90,155,67	92,152,66	92,157,70
    54,104,39	56,104,38	59,106,41	57,105,43	54,104,40	54,103,35	56,101,39	58,99,41	59,99,41
    #background
    114,127,121	117,135,125	120,137,131	132,145,138	142,154,148	151,166,158	160,182,172
    115,125,121	118,131,123	122,132,135	133,142,144	141,151,152	150,166,158	159,179,172

    :param input_file: str
    :param output_file: str
    """
    # If the input file does not exist raise an error
    if not os.path.exists(input_file):
        raise IOError(f"File does not exist: {input_file}")

    # Read the file into a string
    with open(input_file, "r") as fd:
        pixel_data = fd.read()

    # Split the file string by the classname/header character
    classes = pixel_data.split("#")

    # Parse the class data
    class_dict = {}
    # Ignore the first item, it's the empty string or whitespace to the "left" of the first class/header
    for i in range(1, len(classes)):
        # Replace tabs with newlines
        classes[i] = classes[i].replace("\t", "\n")
        # Split the class data on newlines
        class_data = classes[i].split("\n")
        # The class name is the first item
        class_name = class_data[0]
        rgb_values = []
        # Loop over the RGB values, starting with the second element
        for j in range(1, len(class_data)):
            # Skip blank lines but keep the lines with RGB values
            if len(class_data[j]) > 0:
                rgb_values.append(class_data[j])
        class_dict[class_name] = rgb_values

    # Each class could have a different number of RGB values, find the largest
    total_rgb = 0
    for class_name in class_dict:
        if len(class_dict[class_name]) > total_rgb:
            total_rgb = len(class_dict[class_name])

    # Pad the classes with empty strings if they have less than the total RGB values
    for class_name in class_dict:
        missing = total_rgb - len(class_dict[class_name])
        if missing > 0:
            for i in range(missing):
                class_dict[class_name].append("")

    # Open the output file
    with open(output_file, "w") as out:
        # Create the output table
        class_names = class_dict.keys()
        out.write("\t".join(map(str, class_names)) + "\n")
        for i in range(0, total_rgb):
            row = []
            for class_name in class_names:
                row.append(class_dict[class_name][i])
            out.write("\t".join(map(str, row)) + "\n")



"""Color Corrections Functions."""
import os
import math
import cv2
import numpy as np
import altair as alt
import pandas as pd
from plantcv.plantcv import params, fatal_error
from plantcv.plantcv.roi import circle
from plantcv.plantcv._debug import _debug


def affine_color_correction(rgb_img, source_matrix, target_matrix):
    """Affine color correction of RGB image.

    Correct the color of the input image based on the target color matrix using an affine transformation
    in the RGB space. The vector containing the regression coefficients is calculated as the one that minimizes the
    Euclidean distance between the transformed source color values and the target color values.

    Inputs:
    rgb_img         = an RGB image with color chips visualized
    source_matrix   = array of RGB color values (intensity in the range [0-1]) from
                      the image to be corrected where each row is one
                      color reference and the columns are organized as index,R,G,B
    target_matrix   = array of target RGB color values (intensity in the range [0-1])
                      where each row is one color reference and the columns are
                      organized as index,R,G,B

    Outputs:
    corrected_img   = color corrected image


    :param rgb_img: numpy.ndarray
    :return source_matrix: numpy.ndarray
    :return target_matrix: numpy.ndarray
    :return corrected_img: numpy.ndarray
    """
    # matrices must have the same number of color references
    if source_matrix.shape != target_matrix.shape:
        fatal_error("Mismatch between the color matrices' shapes")

    h, w, c = rgb_img.shape

    # number of references
    n = source_matrix.shape[0]

    # the column zero (index) of the matrices is not used in this model
    # augment matrix of source values with a column of 1s for the constant part of
    # the affine transformation
    S = np.concatenate((source_matrix[:, 1:].copy(), np.ones((n, 1))), axis=1)

    # make vectors of target values for each color
    T = target_matrix[:, 1:].copy()
    tr = T[:, 0]
    tg = T[:, 1]
    tb = T[:, 2]

    # calculate regression vector for each color as the pseudoinverse of the source
    # values matrix multiplied by each color target vector
    ar = np.matmul(np.linalg.pinv(S), tr)
    ag = np.matmul(np.linalg.pinv(S), tg)
    ab = np.matmul(np.linalg.pinv(S), tb)

    img_rgb = cv2.cvtColor(rgb_img, cv2.COLOR_BGR2RGB)
    # reshape image as a 2D array where the rows are pixels and the columns are color channels
    # and augment the channels with a column of 1s for the affine transformation
    img_pix = np.concatenate((img_rgb.reshape(h*w, c).astype(np.float64)/255, np.ones((h*w, 1))), axis=1)

    # calculate the corrected colors, eliminate values outside the range [0-1] and
    # convert to [0-255] unit8
    img_r_cc = (255*np.clip(np.matmul(img_pix, ar), 0, 1)).astype(np.uint8)
    img_g_cc = (255*np.clip(np.matmul(img_pix, ag), 0, 1)).astype(np.uint8)
    img_b_cc = (255*np.clip(np.matmul(img_pix, ab), 0, 1)).astype(np.uint8)

    # reconstruct the RGB (actually BGR for openCV) image
    corrected_img = np.stack((img_b_cc, img_g_cc, img_r_cc), axis=1).reshape(h, w, c)

    # For debugging, create a horizontal view of the image before and after color correction
    debug_img = np.hstack([rgb_img, corrected_img])
    _debug(visual=debug_img, filename=os.path.join(params.debug_outdir, str(params.device) + '_affine_corrected.png'))

    return corrected_img


def std_color_matrix(pos=0):
    """Create a standard color matrix.

    Standard color values compatible with the x-rite ColorChecker Classic,
    ColorChecker Mini, and ColorChecker Passport targets.
    Source: https://en.wikipedia.org/wiki/ColorChecker

    Inputs:
    pos     = reference value indicating orientation of the color card. The reference
                is based on the position of the white chip:

                pos = 0: bottom-left corner
                pos = 1: bottom-right corner
                pos = 2: top-right corner
                pos = 3: top-left corner

    Outputs:
    color_matrix    = matrix containing the standard red, green, and blue
                        values for each color chip

    :param pos: int
    :return color_matrix: numpy.ndarray
    """
    # list of rgb values as indicated in the color card specs. They need to be
    # arranged depending on the orientation of the color card of reference in the
    # image to be corrected.
    values_list = np.array([[115, 82, 68],  # dark skin
                            [194, 150, 130],  # light skin
                            [98, 122, 157],  # blue sky
                            [87, 108, 67],  # foliage
                            [133, 128, 177],  # blue flower
                            [103, 189, 170],  # bluish green
                            [214, 126, 44],  # orange
                            [80, 91, 166],  # purplish blue
                            [193, 90, 99],  # moderate red
                            [94, 60, 108],  # purple
                            [157, 188, 64],  # yellow green
                            [224, 163, 46],  # orange yellow
                            [56, 61, 150],  # blue
                            [70, 148, 73],  # green
                            [175, 54, 60],  # red
                            [231, 199, 31],  # yellow
                            [187, 86, 149],  # magenta
                            [8, 133, 161],  # cyan
                            [243, 243, 242],  # white (.05*)
                            [200, 200, 200],  # neutral 8 (.23*)
                            [160, 160, 160],  # neutral 6.5 (.44*)
                            [122, 122, 121],  # neutral 5 (.7*)
                            [85, 85, 85],  # neutral 3.5 (1.05*)
                            [52, 52, 52]], dtype=np.float64)  # black (1.50*)

    pos = math.floor(pos)
    if (pos < 0) or (pos > 3):
        fatal_error("white chip position reference 'pos' must be a value among {0, 1, 2, 3}")

    N_chips = values_list.shape[0]

    # array of indices from 1 to N chips in order to match the chip numbering
    # in the color card specs. Later when used for indexing, we subtract the 1.
    idx = np.arange(N_chips)+1
    # indices in the shape of the color card
    cc_indices = idx.reshape((4, 6), order='C')
    # rotate the indices depending on the specified orientation
    cc_indices_rot = np.rot90(cc_indices, k=pos, axes=(0, 1))
    # arange color values based on the indices
    color_matrix_wo_chip_nb = values_list[(cc_indices_rot-1).reshape(-1), :]/255.

    # add chip number compatible with other PlantCV functions
    chip_nb = np.arange(10, 10*N_chips+1, 10)
    color_matrix = np.concatenate((chip_nb.reshape(N_chips, 1), color_matrix_wo_chip_nb), axis=1)

    return color_matrix


def get_color_matrix(rgb_img, mask):
    """Calculate the average value of pixels in each color chip for each color channel.

    Inputs:
    rgb_img         = RGB image with color chips visualized
    mask        = a gray-scale img with unique values for each segmented space, representing unique, discrete
                    color chips.

    Outputs:
    color_matrix        = a 22x4 matrix containing the average red value, average green value, and average blue value
                            for each color chip.
    headers             = a list of 4 headers corresponding to the 4 columns of color_matrix respectively

    :param rgb_img: numpy.ndarray
    :param mask: numpy.ndarray
    :return headers: string array
    :return color_matrix: numpy.ndarray
    """
    # Check for RGB input
    if len(np.shape(rgb_img)) != 3:
        fatal_error("Input rgb_img is not an RGB image.")
    # Check mask for gray-scale
    if len(np.shape(mask)) != 2:
        fatal_error("Input mask is not an gray-scale image.")

    img_dtype = rgb_img.dtype
    # normalization value as max number if the type is unsigned int
    max_val = 1.0
    if img_dtype.kind == 'u':
        max_val = np.iinfo(img_dtype).max

    # convert to float and normalize to work with values between 0-1
    rgb_img = rgb_img.astype(np.float64)/max_val

    # create empty color_matrix
    color_matrix = np.zeros((len(np.unique(mask))-1, 4))

    # create headers
    headers = ["chip_number", "r_avg", "g_avg", "b_avg"]

    # declare row_counter variable and initialize to 0
    row_counter = 0

    # for each unique color chip calculate each average RGB value
    for i in np.unique(mask):
        if i != 0:
            chip = rgb_img[np.where(mask == i)]
            color_matrix[row_counter][0] = i
            color_matrix[row_counter][1] = np.mean(chip[:, 2])
            color_matrix[row_counter][2] = np.mean(chip[:, 1])
            color_matrix[row_counter][3] = np.mean(chip[:, 0])
            row_counter += 1

    return headers, color_matrix


def get_matrix_m(target_matrix, source_matrix):
    """Calculate Moore-Penrose inverse matrix for use in calculating transformation_matrix.

    Inputs:
    target_matrix       = a 22x4 matrix containing the average red value, average green value, and average blue value
                            for each color chip.
    source_matrix       = a 22x4 matrix containing the average red value, average green value, and average blue value
                            for each color chip.

    Outputs:
    matrix_a    = a concatenated 22x9 matrix of source_matrix red, green, and blue values to the powers 1, 2, 3
    matrix_m    = a 9x22 Moore-Penrose inverse matrix
    matrix_b    = a 22x9 matrix of linear, square, and cubic rgb values from target_img


    :param target_matrix: numpy.ndarray
    :param source_matrix: numpy.ndarray
    :return matrix_a: numpy.ndarray
    :return matrix_m: numpy.ndarray
    :return matrix_b: numpy.ndarray
    """
    # if the number of chips in source_img match the number of chips in target_matrix
    if np.shape(target_matrix) == np.shape(source_matrix):
        t_mats = np.split(target_matrix, 4, 1)
        t_r, t_g, t_b = t_mats[1], t_mats[2], t_mats[3]
        s_mats = np.split(source_matrix, 4, 1)
        s_r, s_g, s_b = s_mats[1], s_mats[2], s_mats[3]
    else:
        combined_matrix = np.zeros((np.ma.size(source_matrix, 0), 7))
        row_count = 0
        for r in range(0, np.ma.size(target_matrix, 0)):
            for i in range(0, np.ma.size(source_matrix, 0)):
                if target_matrix[r][0] == source_matrix[i][0]:
                    combined_matrix[row_count][0] = target_matrix[r][0]
                    combined_matrix[row_count][1] = target_matrix[r][1]
                    combined_matrix[row_count][2] = target_matrix[r][2]
                    combined_matrix[row_count][3] = target_matrix[r][3]
                    combined_matrix[row_count][4] = source_matrix[i][1]
                    combined_matrix[row_count][5] = source_matrix[i][2]
                    combined_matrix[row_count][6] = source_matrix[i][3]
                    row_count += 1
        rgb_mats = np.split(combined_matrix, 7, 1)
        t_r, t_g, t_b, s_r, s_g, s_b = rgb_mats[1], rgb_mats[2], rgb_mats[3], rgb_mats[4], rgb_mats[5], rgb_mats[6]
    t_r2 = np.square(t_r)
    t_r3 = np.power(t_r, 3)
    t_g2 = np.square(t_g)
    t_g3 = np.power(t_g, 3)
    t_b2 = np.square(t_b)
    t_b3 = np.power(t_b, 3)
    s_r2 = np.square(s_r)
    s_r3 = np.power(s_r, 3)
    s_g2 = np.square(s_g)
    s_g3 = np.power(s_g, 3)
    s_b2 = np.square(s_b)
    s_b3 = np.power(s_b, 3)

    # create matrix_a
    matrix_a = np.concatenate((s_r, s_g, s_b, s_r2, s_g2, s_b2, s_r3, s_g3, s_b3), 1)
    # create matrix_m
    matrix_m = np.linalg.solve(np.matmul(matrix_a.T, matrix_a), matrix_a.T)
    # create matrix_b
    matrix_b = np.concatenate((t_r, t_r2, t_r3, t_g, t_g2, t_g3, t_b, t_b2, t_b3), 1)
    return matrix_a, matrix_m, matrix_b


def calc_transformation_matrix(matrix_m, matrix_b):
    """Calculate a transformation matrix (transformation_matrix).

    Inputs:
    matrix_m    = a 9x22 Moore-Penrose inverse matrix
    matrix_b    = a 22x9 matrix of linear, square, and cubic rgb values from target_img

    Outputs:
    1-t_det     = "deviance" the measure of how greatly the source image deviates from the target image's color space.
                    Two images of the same color space should have a deviance of ~0.
    transformation_matrix    = a 9x9 matrix of linear, square, and cubic transformation coefficients


    :param matrix_m: numpy.ndarray
    :param matrix_b: numpy.ndarray
    :return red: numpy.ndarray
    :return blue: numpy.ndarray
    :return green: numpy.ndarray
    :return 1-t_det: float
    :return transformation_matrix: numpy.ndarray
    """
    # check matrix_m and matrix_b are matrices
    if len(np.shape(matrix_b)) != 2 or len(np.shape(matrix_m)) != 2:
        fatal_error("matrix_m and matrix_b must be n x m matrices such that m,n != 1.")
    # check matrix_b has 9 columns
    if np.shape(matrix_b)[1] != 9:
        fatal_error("matrix_b must have 9 columns.")
    # check matrix_m and matrix_b for multiplication
    if np.shape(matrix_m)[0] != np.shape(matrix_b)[1] or np.shape(matrix_m)[1] != np.shape(matrix_b)[0]:
        fatal_error("Cannot multiply matrices.")

    t_mats = np.split(matrix_b, 9, 1)
    t_r, t_r2, t_r3 = t_mats[0], t_mats[1], t_mats[2]
    t_g, t_g2, t_g3 = t_mats[3], t_mats[4], t_mats[5]
    t_b, t_b2, t_b3 = t_mats[6], t_mats[7], t_mats[8]
    # multiply each 22x1 matrix from target color space by matrix_m
    red = np.matmul(matrix_m, t_r)
    green = np.matmul(matrix_m, t_g)
    blue = np.matmul(matrix_m, t_b)

    red2 = np.matmul(matrix_m, t_r2)
    green2 = np.matmul(matrix_m, t_g2)
    blue2 = np.matmul(matrix_m, t_b2)

    red3 = np.matmul(matrix_m, t_r3)
    green3 = np.matmul(matrix_m, t_g3)
    blue3 = np.matmul(matrix_m, t_b3)

    # concatenate each product column into 9X9 transformation matrix
    transformation_matrix = np.concatenate((red, green, blue, red2, green2, blue2, red3, green3, blue3), 1)

    # find determinant of transformation matrix
    t_det = np.linalg.det(transformation_matrix)

    return 1-t_det, transformation_matrix


def apply_transformation_matrix(source_img, target_img, transformation_matrix):
    """Apply the transformation matrix to the source_image.

    Inputs:
    source_img      = an RGB image to be corrected to the target color space
    target_img      = an RGB image with the target color space
    transformation_matrix        = a 9x9 matrix of transformation coefficients

    Outputs:
    corrected_img    = an RGB image in correct color space

    :param source_img: numpy.ndarray
    :param target_img: numpy.ndarray
    :param transformation_matrix: numpy.ndarray
    :return corrected_img: numpy.ndarray
    """
    # check transformation_matrix for 9x9
    if np.shape(transformation_matrix) != (9, 9):
        fatal_error("transformation_matrix must be a 9x9 matrix of transformation coefficients.")
    # Check for RGB input
    if len(np.shape(source_img)) != 3:
        fatal_error("Source_img is not an RGB image.")

    # split transformation_matrix
    channels = np.split(transformation_matrix, 9, 1)
    red, green, blue = channels[0], channels[1], channels[2]

    source_dtype = source_img.dtype
    # normalization value as max number if the type is unsigned int
    max_val = 1.0
    if source_dtype.kind == 'u':
        max_val = np.iinfo(source_dtype).max
    # convert img to float to avoid integer overflow, normalize between 0-1
    source_flt = source_img.astype(np.float64)/max_val
    # find linear, square, and cubic values of source_img color channels
    source_b, source_g, source_r = cv2.split(source_flt)
    source_b2 = np.square(source_b)
    source_b3 = np.power(source_b, 3)
    source_g2 = np.square(source_g)
    source_g3 = np.power(source_g, 3)
    source_r2 = np.square(source_r)
    source_r3 = np.power(source_r, 3)

    # apply linear model to source color channels
    b = 0 + source_r * blue[0] + source_g * blue[1] + source_b * blue[2] + source_r2 * blue[3] + source_g2 * blue[
        4] + source_b2 * blue[5] + source_r3 * blue[6] + source_g3 * blue[7] + source_b3 * blue[8]
    g = 0 + source_r * green[0] + source_g * green[1] + source_b * green[2] + source_r2 * green[3] + source_g2 * green[
        4] + source_b2 * green[5] + source_r3 * green[6] + source_g3 * green[7] + source_b3 * green[8]
    r = 0 + source_r * red[0] + source_g * red[1] + source_b * red[2] + source_r2 * red[3] + source_g2 * red[
        4] + source_b2 * red[5] + source_r3 * red[6] + source_g3 * red[7] + source_b3 * red[8]

    # merge corrected color channels onto source_image
    bgr = [b, g, r]
    corrected_img = cv2.merge(bgr)

    # return values of the image to the original range
    corrected_img = max_val*np.clip(corrected_img, 0, 1)
    # cast back to original dtype (if uint the value defaults to the closest smaller integer)
    corrected_img = corrected_img.astype(source_dtype)

    # For debugging, create a horizontal view of source_img, corrected_img, and target_img to the plotting device
    # plot horizontal comparison of source_img, corrected_img (with rounded elements) and target_img
    out_img = np.hstack([source_img, corrected_img, target_img])
    # Change range of visualization image to 0-255 and convert to uin8
    out_img = ((255.0/max_val)*out_img).astype(np.uint8)
    _debug(visual=out_img, filename=os.path.join(params.debug_outdir, str(params.device) + '_corrected.png'))

    # return the corrected image
    return corrected_img


def save_matrix(matrix, filename):
    """Serialize a matrix as an numpy.ndarray object and save to a .npz file.

    Inputs:
    matrix      = a numpy.matrix
    filename    = name of file to which matrix will be saved. Must end in .npz

    :param matrix: numpy.ndarray
    :param filename: string ending in ".npz"
    """
    if ".npz" not in filename:
        fatal_error("File must be an .npz file.")

    # Autoincrement the device counter
    params.device += 1

    np.savez(filename, matrix)


def load_matrix(filename):
    """Deserializes from file an numpy.ndarray object as a matrix.

    Inputs:
    filename    = .npz file to which a numpy.matrix or numpy.ndarray is saved

    Outputs:
    matrix      = a numpy.matrix

    :param filename: string ending in ".npz"
    :return matrix: numpy.matrix
    """
    matrix_file = np.load(filename, encoding="latin1")
    matrix = matrix_file['arr_0']
    np.asmatrix(matrix)

    return matrix


def correct_color(target_img, target_mask, source_img, source_mask, output_directory):
    """Take a target_img with preferred color_space and convert source_img to that color_space.

    Inputs:
    target_img          = an RGB image with color chips visualized
    source_img          = an RGB image with color chips visualized
    target_mask         = a gray-scale image with color chips and background each represented with unique values
    target_mask         = a gray-scale image with color chips and background each represented as unique values
    output_directory    = a file path to which outputs will be saved
    Outputs:
    target_matrix   = saved in .npz file, a 22x4 matrix containing the average red value, average green value, and
                            average blue value for each color chip.
    source_matrix   = saved in .npz file, a 22x4 matrix containing the average red value, average green value, and
                            average blue value for each color chip.
    transformation_matrix        = saved in .npz file, a 9x9 transformation matrix

    corrected_img   = the source_img converted to the correct color space.


    :param target_img: numpy.ndarray
    :param source_img: numpy.ndarray
    :param target_mask: numpy.ndarray
    :param source_mask: numpy.ndarray
    :param output_directory: string
    :return target_matrix: numpy.matrix
    :return source_matrix: numpy.matrix
    :return transformation_matrix: numpy.matrix
    :return corrected_img: numpy.ndarray
    """
    # check output_directory, if it does not exist, create
    if not os.path.exists(output_directory):
        os.mkdir(output_directory)

    # get color matrices for target and source images
    _, target_matrix = get_color_matrix(target_img, target_mask)
    _, source_matrix = get_color_matrix(source_img, source_mask)

    # save target and source matrices
    save_matrix(target_matrix, os.path.join(output_directory, "target_matrix.npz"))
    save_matrix(source_matrix, os.path.join(output_directory, "source_matrix.npz"))

    # get matrix_m
    _, matrix_m, matrix_b = get_matrix_m(target_matrix=target_matrix, source_matrix=source_matrix)
    # calculate transformation_matrix and save
    _, transformation_matrix = calc_transformation_matrix(matrix_m, matrix_b)
    save_matrix(transformation_matrix, os.path.join(output_directory, "transformation_matrix.npz"))

    # apply transformation
    corrected_img = apply_transformation_matrix(source_img, target_img, transformation_matrix)

    return target_matrix, source_matrix, transformation_matrix, corrected_img


def create_color_card_mask(rgb_img, radius, start_coord, spacing, nrows, ncols, exclude=None):
    """Create a labeled mask for color card chips.

    Inputs:
    rgb_img        = Input RGB image data containing a color card.
    radius         = Radius of color masks.
    start_coord    = Two-element tuple of the first chip mask starting x and y coordinate.
    spacing        = Two-element tuple of the horizontal and vertical spacing between chip masks.
    nrows          = Number of chip rows.
    ncols          = Number of chip columns.
    exclude        = Optional list of chips to exclude.

    Returns:
    mask           = Labeled mask of chips

    :param rgb_img: numpy.ndarray
    :param radius: int
    :param start_coord: tuple
    :param spacing: tuple
    :param nrows: int
    :param ncols: int
    :param exclude: list
    :return mask: numpy.ndarray
    """
    # Initialize chip list
    chips = []
    # Store debug mode
    debug = params.debug
    params.debug = None
    # Initialize exclude list
    if exclude is None:
        exclude = []
    # Loop over each color card row
    for i in range(0, nrows):
        # The upper left corner is the y starting coordinate + the chip offset * the vertical spacing between chips
        y = start_coord[1] + i * spacing[1]
        # Loop over each column
        for j in range(0, ncols):
            # The upper left corner is the x starting coordinate + the chip offset * the
            # horizontal spacing between chips
            x = start_coord[0] + j * spacing[0]
            # Create a chip ROI
            chips.append(circle(img=rgb_img, x=x, y=y, r=radius))
    # Restore debug parameter
    params.debug = debug
    # Sort excluded chips from largest to smallest
    exclude.sort(reverse=True)
    # Remove any excluded chips
    for chip in exclude:
        del chips[chip]
    # Create mask
    mask = np.zeros(shape=np.shape(rgb_img)[:2], dtype=np.uint8)
    # Mask label index
    i = 1
    # Draw labeled chip boxes on the mask
    for chip in chips:
        mask = cv2.drawContours(mask, chip.contours[0], -1, (i * 10), -1)
        i += 1
    # Create a copy of the input image for plotting
    canvas = np.copy(rgb_img)
    # Draw chip ROIs on the canvas image
    for chip in chips:
        cv2.drawContours(canvas, chip.contours[0], -1, (255, 255, 0), params.line_thickness)
    _debug(visual=canvas, filename=os.path.join(params.debug_outdir, str(params.device) + '_color_card_mask_rois.png'))
    _debug(visual=mask, filename=os.path.join(params.debug_outdir, str(params.device) + '_color_card_mask.png'))
    return mask


def quick_color_check(target_matrix, source_matrix, num_chips):
    """Plot the color values of a target and source color card matrix.

    Quickly plot target matrix values against source matrix values to determine
    over saturated color chips or other issues.

    Inputs:
    source_matrix      = an nrowsXncols matrix containing the avg red, green, and blue values for each color chip
                            of the source image
    target_matrix      = an nrowsXncols matrix containing the avg red, green, and blue values for each color chip
                            of the target image
    num_chips          = number of color card chips included in the matrices (integer)

    Returns:
    p1                 = an altair plot of the target and source color values

    :param source_matrix: numpy.ndarray
    :param target_matrix: numpy.ndarray
    :param num_chips: int
    :return p1: altair.vegalite.v5.api.Chart
    """
    # Scale matrices to 0-255
    target_matrix = 255*target_matrix
    source_matrix = 255*source_matrix

    # Extract and organize matrix info
    tr = target_matrix[:num_chips, 1:2]
    tg = target_matrix[:num_chips, 2:3]
    tb = target_matrix[:num_chips, 3:4]
    sr = source_matrix[:num_chips, 1:2]
    sg = source_matrix[:num_chips, 2:3]
    sb = source_matrix[:num_chips, 3:4]

    # Create columns of color labels
    red = ["red"] * num_chips
    blue = ["blue"] * num_chips
    green = ["green"] * num_chips

    # Make a column of chip numbers
    chip = np.arange(0, num_chips).reshape((num_chips, 1))
    chips = np.vstack((chip, chip, chip))

    # Combine info
    color_data_r = np.column_stack((sr, tr, red))
    color_data_g = np.column_stack((sg, tg, green))
    color_data_b = np.column_stack((sb, tb, blue))
    all_color_data = np.vstack((color_data_b, color_data_g, color_data_r))

    # Create a dataframe with headers
    dataset = pd.DataFrame({'source': all_color_data[:, 0], 'target': all_color_data[:, 1],
                            'color': all_color_data[:, 2]})

    # Add chip numbers to the dataframe
    dataset['chip'] = chips
    dataset = dataset.astype({'color': str, 'chip': str, 'target': float, 'source': float})

    # Make the plot
    p1 = alt.Chart(dataset).mark_point(point=True).encode(
        x="target",
        y="source",
        color=alt.Color("color").scale(range=["blue", "green", "red"]),
        column="color"
        )

    _debug(visual=p1, filename=os.path.join(params.debug_outdir, 'color_quick_check.png'))

    return p1.interactive()



import pandas as pd
import os
from plantcv.parallel.parsers import metadata_parser
from plantcv.parallel.workflowconfig import WorkflowConfig


def inspect_dataset(config):
    """Inspect a dataset before running plantcv in parallel over the directory
    Parameters
    ----------
    config   = plantcv.parallel.WorkflowConfig object or str, if str then very minimal processing is done

    Returns
    -------
    summary_df     = pandas.core.frame.DataFrame,
        dataframe of number of unique values of metadata filters and statuses.
    meta     = pandas.core.frame.DataFrame, dataframe of image metadata.
    """
    if isinstance(config, str):
        input_dir = config
        config = WorkflowConfig()
        if input_dir.endswith(".json"):
            config.import_config(config_file=input_dir)
        else:
            config.input_dir = input_dir
    # run the metadata parser to find images and return dataframes
    meta, removed = metadata_parser(config)
    # make dataframe out of groupby object
    meta_filepaths = []
    for i, _ in meta["filepath"]:
        meta_filepaths.append(i[0])
    meta = meta.apply(lambda x: x, include_groups=False)
    meta["filepath"] = meta_filepaths
    # flag kept images
    meta["status"] = "Kept"
    # combine both dataframes
    if not removed.empty:
        meta = pd.concat(df.dropna(axis=1, how='all') for df in [meta, removed])
    # pull extensions
    meta['extension'] = meta['filepath'].apply(
        lambda x: os.path.splitext(os.path.basename(x))[1]
    )
    # count unique values by all filtered metadata and steps where images were dropped
    summary_df = meta.groupby(
        ['status', *config.metadata_filters],
        dropna=False
    ).agg(_agg_unique_values)
    return summary_df, meta


def _agg_unique_values(series):
    """Aggregate columns to a number of unique values and possibly their values
    Parameters
    ----------
    series     = pandas.core.series.Series, a metadata column of a pandas dataframe

    Return
    ------
    label      = str, a string describing the number of unique values and their
                 observed levels if there are <4.
    """
    n_unique = series.nunique()
    context = ""
    if n_unique:
        context = " (...)"
    series = series.astype("string")
    # if there are only a few options then print what they are
    if n_unique and n_unique < 4:
        uniques = series.dropna().unique()
        unique_vals = ", ".join(uniques)
        context = " (" + unique_vals + ")"
    label = str(n_unique) + context
    return label



# RGB -> Gray

import os
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2gray


def rgb2gray(rgb_img):
    """Convert image from RGB colorspace to Gray.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        RGB image data

    Returns
    -------
    numpy.ndarray
        grayscale image
    """
    gray = _rgb2gray(rgb_img=rgb_img)

    _debug(visual=gray, filename=os.path.join(params.debug_outdir, f"{params.device}_gray.png"))

    return gray



"""Reassign PSII frame labels based on induction curve"""
import numpy as np
from plantcv.plantcv import fatal_error
from plantcv.plantcv.classes import PSII_data
from plantcv.plantcv import params


def reassign_frame_labels(ps_da, mask):
    """
    Analyze fluorescence induction curve and assign Fm or Fmp frame labels.

    Designed for cropreporter data. Analyze fluorescence frames to find max mean fluorescence and assign Fm or Fmp.
    Use this if you want to assign Fm/Fmp based on observed values rather than CropReporter metadata.

    Inputs:
    ps_da       = photosynthesis xarray DataArray
    mask        = mask of plant (binary, single channel)

    Returns:
    ps_da       = dataarray with updated frame_label coordinate
    ind_fig     = ggplot induction curve of fluorescence
    ind_df      = data frame of mean fluorescence in the masked region at each timepoint

    :param ps_da: xarray.core.dataarray.DataArray
    :param mask: numpy.ndarray
    :return ps_da: xarray.core.dataarray.DataArray
    :return ind_fig: ggplot figure
    :return ind_df: pandas.core.frame.DataFrame
    """
    params.device += 1

    try:
        if ps_da.name not in ["ojip_light", "ojip_dark"]:
            fatal_error("You must provide a xarray DataArray with name ojip_light or ojip_dark")
    except AttributeError:
        if isinstance(ps_da, PSII_data):
            fatal_error("You need to provide the `ojip_dark` or `ojip_light` dataarray")
        else:
            fatal_error("You must provide a xarray DataArray with name ojip_light or ojip_dark")

    if mask.shape != ps_da.shape[:2] or len(np.unique(mask)) > 2:
        fatal_error(f"Mask needs to be binary and have shape {ps_da.shape[:2]}")

    # Prime is empty for Fv/Fm (dark- and light-adapted) and p for Fq'/Fm'
    datasets = {
        "ojip_light": {
            "prime": "p",
            "label": "PSL",
            "F": "Fp"
        },
        "ojip_dark": {
            "prime": "",
            "label": "PSD",
            "F": "F0"
        }
    }

    # Get the number of frame labels
    ind_size = ps_da.frame_label.size
    # Create a new frame label array populated with the current labels
    idx = ps_da.frame_label.values
    # Find the frame corresponding to the first frame after F0/Fp
    f = idx.tolist().index(datasets[ps_da.name.lower()]['F']) + 1
    # Reset the frame labels after F0/Fp
    for i in range(f, ind_size):
        idx[i] = f"{datasets[ps_da.name.lower()]['label']}{i}"
    # get plant mean for each frame based on mask
    exp_mask = np.copy(mask)[..., None, None]
    fluor_values = ps_da.where(exp_mask > 0).mean(['x', 'y', 'measurement'])
    # find frame with max mean after the control and F/F' frames
    max_ind = np.argmax(fluor_values.data[f:])
    # assign max frame label
    idx[max_ind + f] = f"Fm{datasets[ps_da.name.lower()]['prime']}"
    # assign new labels back to dataarray
    ps_da = ps_da.assign_coords({'frame_label': ('frame_label', idx)})

    return ps_da



"""Write hyperspectral image data to a file."""
import os
from plantcv.plantcv import __version__ as version


def write_data(filename, spectral_data):
    """Write hyperspectral image data to a file.
    Inputs:
    filename        = Name of file to write
    spectral_data   = Hyperspectral data object

    Returns:

    :param filename: str
    :param spectral_data: __main__.Spectral_data
    """
    filename = os.path.splitext(filename)[0]

    # create header
    lines, samples, bands = spectral_data.array_data.shape
    dtype_dict = {'B': "1", 'h': "2", 'i': "3", 'f': "4", 'd': "5", 'F': "6",
                  'D': "9", 'H': "12", 'I': "13", 'l': "14", 'L': "15"}
    wavelenghths = list(spectral_data.wavelength_dict.keys())
    with open(filename+'.hdr', mode='w') as f:
        f.write('ENVI\n')
        f.write(f'; this file was created using PlantCV version {version}\n')
        f.write(f'; original file: {spectral_data.filename}\n')
        f.write('interleave = bil\n')
        f.write(f'samples = {samples}\n')
        f.write(f'lines = {lines}\n')
        f.write(f'bands = {bands}\n')
        f.write(f'data type = {dtype_dict[spectral_data.array_data.dtype.char]}\n')
        f.write(f'wavelength units = {spectral_data.wavelength_units}\n')
        if spectral_data.default_bands is not None:
            db_string = ",".join([str(band) for band in spectral_data.default_bands])
            f.write(f'default bands = {{{db_string}}}\n')
        f.write('wavelength = {\n')
        for wl in wavelenghths[:-1]:
            f.write(f'{wl},\n')
        f.write(f'{wavelenghths[-1]}\n')
        f.write('}')

    # create raw binary file containing the hyperspectral array values
    with open(filename+'.raw', mode='w+b') as f:
        f.write(spectral_data.array_data.transpose(0, 2, 1).tobytes(order='C'))



"""Open an image from a URL."""
import imageio.v3 as iio
import cv2
from plantcv.plantcv import fatal_error
from plantcv.plantcv._debug import _debug


def open_url(url):
    """Open an image from a URL and return it as a numpy array.

    Parameters
    ----------
    url : str
        URL of the image to be opened.

    Returns
    -------
    numpy.ndarray
        Image data as a numpy array.
    """
    # Read the image from the URL using imageio
    image = iio.imread(url)

    # Check if the image is grayscale or RGB
    if len(image.shape) not in [2, 3]:
        fatal_error("Image is not RGB or grayscale.")

    if image.shape[-1] == 3:
        # Convert the image to RGB format
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

    # Debugging visualization
    _debug(visual=image, filename="url_image.png")

    return image



# merging images with a known overlap

import cv2
import os
import numpy as np
import random
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug


def merge_images(paths_to_imgs, overlap_percentage, direction="vertical", method="stacked"):
    """
    Merge together images in a series that overlap by a specified amount.
    Inputs:
    paths_to_imgs = List of paths to the images
    overlap_percentage = percent of each image that overlaps with adjacent images
    direction = Available options are vertical or horizontal and indicate
        how the images should be merged
        - 'horizontal' : merge images on the x-axis
        - 'vertical' : (default) merge images on the y-axis
    method = Available options are stacked, random, average, and gradual and dictate
        how the overlap region should be handled
        - 'stacked' : (default), image i+1 stacks on top of image i
        - 'random' : randomly choose either image i or i+1 to go on top
        - 'average' : pixels are averaged between image i values and image i+1 values
        - 'gradual' : pixels are averaged with a weight that corresponds to
                      proximity to image i or image i+1
    :param paths_to_imgs: list
    :param overlap_percentage: non-negative real number
    :param direction: str
    :param method: str
    :return combined_image: numpy.ndarray
    """
    paths_to_imgs.sort()

    # Read the images to get total height/width
    overlap_dims = [0, 0]
    for imgfile in paths_to_imgs:
        height, width, _ = cv2.imread(imgfile).shape
        overlap_dims[0] += height
        overlap_dims[1] += width
    height, width, _ = cv2.imread(paths_to_imgs[0]).shape

    # Calculate the overlap in pixels
    # Create a new image with adjusted dimensions accounting for overlap
    if direction == "vertical":
        overlap_pixels_height = int(height * overlap_percentage / 100)
        combined_height = overlap_dims[0] - (overlap_pixels_height*(len(paths_to_imgs)-1))
        blank_image = np.zeros((combined_height, width, 3), dtype=np.uint8)
        combined_image = _mergevert(paths_to_imgs, blank_image, height, overlap_pixels_height, method)

    elif direction == "horizontal":
        overlap_pixels_width = int(width * overlap_percentage / 100)
        combined_width = overlap_dims[1] - (overlap_pixels_width*(len(paths_to_imgs)-1))
        blank_image = np.zeros((height, combined_width, 3), dtype=np.uint8)
        combined_image = _mergehoriz(paths_to_imgs, blank_image, width, overlap_pixels_width, method)

    _debug(visual=combined_image, filename=os.path.join(params.debug_outdir, "_merged_image.png"))
    return combined_image


def _mergevert(image_files, combined_image, height, overlap_pixels, method="stacked"):
    """
    Private function to reduce if/else statements. Merges images vertically.
    Inputs: same as above with the addition of calculated overlap pixels and
            dimensions of combined image.
    """
    top = 0  # Setting a dummy variable in case method not random
    for i, image_file in enumerate(image_files):
        image = cv2.imread(image_file)
        if i == 0:  # First image
            combined_image[:height, :] = image
        else:
            start_x = i * (height - overlap_pixels)
            if method == "random":
                top = random.choice([0, 1])
            # Blend the overlap by averaging the pixel values
            for x in range(overlap_pixels):
                betalist = [["stacked", "average", "gradual", "random"], [1, 0.5, x/overlap_pixels, top]]
                beta = betalist[1][betalist[0].index(method)]
                alpha = 1 - beta
                merged_x = start_x + x
                combined_image[merged_x, :] = cv2.addWeighted(combined_image[merged_x, :], alpha,
                                                              image[x, :], beta, 0)
            # Copy the non-overlapping part
            non_overlap_start = start_x + overlap_pixels
            non_overlap_end = non_overlap_start + height - overlap_pixels
            combined_image[non_overlap_start:non_overlap_end, :] = image[overlap_pixels:, :]
    return combined_image


def _mergehoriz(image_files, combined_image, width, overlap_pixels, method="stacked"):
    """
    Private function to reduce if/else statements. Merges images horizontally.
    Inputs: same as above with the addition of calculated overlap pixels and
            dimensions of combined image.
    """
    top = 0  # Setting a dummy variable in case method not random
    for i, image_file in enumerate(image_files):
        image = cv2.imread(image_file)
        if i == 0:  # First image
            combined_image[:, :width] = image
        else:
            start_x = i * (width - overlap_pixels)
            if method == "random":
                top = random.choice([0, 1])
            # Blend the overlap by averaging the pixel values
            for x in range(overlap_pixels):
                betalist = [["stacked", "average", "gradual", "random"], [1, 0.5, x/overlap_pixels, top]]
                beta = betalist[1][betalist[0].index(method)]
                alpha = 1 - beta
                merged_x = start_x + x
                combined_image[:, merged_x] = cv2.addWeighted(combined_image[:, merged_x], alpha,
                                                              image[:, x], beta, 0)
            # Copy the non-overlapping part
            non_overlap_start = start_x + overlap_pixels
            non_overlap_end = non_overlap_start + width - overlap_pixels
            combined_image[:, non_overlap_start:non_overlap_end] = image[:, overlap_pixels:]
    return combined_image



# camera calibration function using checkerboard images

import cv2 as cv
import os
import numpy as np
from plantcv.plantcv.readimage import readimage
from plantcv.plantcv import params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv.transform.color_correction import save_matrix
from plantcv.plantcv.transform.color_correction import load_matrix
from plantcv.plantcv._helpers import _rgb2gray


def checkerboard_calib(img_path, col_corners, row_corners, out_dir):
    """
    Use several checkerboard images to calibrate a camera with image distortions.
    Inputs:
    img_path    = directory of checkerboard images to be used for calibration
    col_corners = the number from inside corners in a column of the checkerboard
    row_corners = the number from inside corners in a row of the checkerboard
    output_directory = filepath where the outputs will be saved

    :param img_path: path to directory of checkerboard images
    :param col_corners: non-negative real number
    :param row_corners: non-negative real number
    :param output_directory = string
    :return mtx: numpy.ndarray
    :return dist: numpy.ndarray
    """
    images = os.listdir(img_path)
    objp = np.zeros((col_corners*row_corners, 3), np.float32)
    objp[:, :2] = np.mgrid[0:col_corners, 0:row_corners].T.reshape(-1, 2)
    # Arrays to store object points and image points from all the images.
    objpoints = []  # 3d point in real world space
    imgpoints = []  # 2d points in image plane

    for fname in images:
        img, _, _ = readimage(filename=os.path.join(img_path, fname), mode="native")
        img1 = np.copy(img)
        gray_img = _rgb2gray(img1)
        ret, corners = cv.findChessboardCorners(gray_img, (col_corners, row_corners))
        criteria = (cv.TERM_CRITERIA_EPS + cv.TERM_CRITERIA_MAX_ITER, 30, 0.001)
        if ret is True:
            objpoints.append(objp)
            corners2 = cv.cornerSubPix(gray_img, corners, (11, 11), (-1, -1), criteria)
            imgpoints.append(corners2)
            # Draw and display the corners
            debug_mode = params.debug
            params.debug = None
            out_img = cv.drawChessboardCorners(img1, (col_corners, row_corners), corners2, ret)
            # Debug images
            params.debug = debug_mode
            _debug(visual=out_img, filename=os.path.join(params.debug_outdir, str(params.device) +
                                                         "_checkerboard_corners.png"))
        else:
            print("Checkerboard image " + fname + " does not match given dimensions.")

    _, mtx, dist, _, _ = cv.calibrateCamera(objpoints, imgpoints, gray_img.shape[::-1], None, None)

    # check output_directory, if it does not exist, create
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    # save matrices
    save_matrix(mtx, os.path.join(out_dir, "mtx.npz"))
    save_matrix(dist, os.path.join(out_dir, "dist.npz"))

    return mtx, dist


def calibrate_camera(rgb_img, mtx_filename, dist_filename):
    """
    Use the outputs from checkerboard_calib to correct the distortions in an image
    Inputs:
    img  = an RGB image
    mtx  = a .npz file, an output of checkerboar_calib
    dist = a .npz file, an output of checkerboar_calib

    :param img: path to an image
    :param ret: float
    :param mtx: numpy.ndarray
    :param dist: numpy.ndarray
    :return corrected_img: numpy.ndarray
    """
    mtx = load_matrix(mtx_filename)
    dist = load_matrix(dist_filename)

    h, w = rgb_img.shape[:2]

    newcameramtx, _ = cv.getOptimalNewCameraMatrix(mtx, dist, (w, h), 1, (w, h))

    corrected_img = cv.undistort(rgb_img, mtx, dist, None, newcameramtx)

    # Debug images
    _debug(visual=corrected_img, filename=os.path.join(params.debug_outdir, str(params.device) + "_checkerboard_corners.png"))

    return corrected_img



# Random Subset a list

import random
from plantcv.plantcv import fatal_error


def random_subset(dataset, num=100, seed=None):
    """
    Get a random subset of the elements in a list.
    Inputs:
    dataset      = List to be subset
    num          = Number of elements in the resulting subset
    seed         = Optional seed for the random number generator
    Returns:
    sub_dataset = List of 'num' elements randomly drawn from 'dataset'
    :param dataset: list
    :param num: int
    :param seed: int
    :return sub_dataset: list
    """
    random.seed(a=seed)

    # Check to make sure number of imgs to select is less than number of images found
    N = len(dataset)
    if num > N:
        fatal_error("Number of images found less than 'num'.")

    # Get random images
    samples = random.sample(dataset, k=num)

    return samples



# Gaussian blur device

import cv2
import os
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import params


def gaussian_blur(img, ksize, sigma_x=0, sigma_y=None, roi=None):
    """
    Applies a Gaussian blur filter.

    Inputs:
    # img     = RGB or grayscale image data
    # ksize   = Tuple of kernel dimensions, e.g. (5, 5)
    # sigmax  = standard deviation in X direction; if 0, calculated from kernel size
    # sigmay  = standard deviation in Y direction; if sigmaY is None, sigmaY is taken to equal sigmaX
    # roi     = Optional rectangular ROI to apply gaussian blur in

    Returns:
    img_gblur = blurred image

    :param img: numpy.ndarray
    :param ksize: tuple
    :param sigma_x: int
    :param sigma_y: str or int
    :param roi: plantcv.plantcv.Objects
    :return img_gblur: numpy.ndarray
    """
    sub_img_gblur = _rect_filter(img, roi, cv2.GaussianBlur,
                                 **{"ksize": ksize, "sigmaX": sigma_x, "sigmaY": sigma_y})
    img_gblur = _rect_replace(img, sub_img_gblur, roi)
    if len(np.shape(img_gblur)) == 3:
        cmap = None
    else:
        cmap = 'gray'

    _debug(visual=img_gblur,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_gaussian_blur.png'),
           cmap=cmap)

    return img_gblur



# Plot segment ID numbers after combining segments

import os
import cv2
import numpy as np
from plantcv.plantcv import params
from plantcv.plantcv import fatal_error
from plantcv.plantcv import color_palette
from plantcv.plantcv._debug import _debug


def segment_combine(segment_list, objects, mask):
    """Combine user specified segments together.

    Inputs:
    segment_list  = List of segment indices to get combined
    objects       = List of contours
    mask          = Binary mask for debugging image

    Returns:
    segmented_img = Segmented image
    objects       = Updated list of contours

    :param segment_list: list
    :param objects: list
    :param mask: numpy.ndarray
    :return labeled_img: numpy.ndarray
    :return objects: list
    """
    label_coord_x = []
    label_coord_y = []
    all_objects = objects[:]
    if type(segment_list[0]) is not int:
        fatal_error("segment_list must be a list of object ID's")
    segment_list_copy = sorted(segment_list, reverse=True)

    # If user provides a single list of objects to combine
    num_contours = len(segment_list)
    count = 1

    # Store the first object into the new object array
    combined_object = objects[segment_list_copy[0]]
    # Remove the objects getting combined from the list of all objects
    all_objects.pop(segment_list_copy[0])

    while count < num_contours:
        # Combine segments into a single object
        combined_object = np.append(combined_object, objects[segment_list_copy[count]], 0)
        # Remove the segment that was combined from the list of all objects
        all_objects.pop(segment_list_copy[count])
        count += 1
    # Replace with the combined object
    all_objects.append(combined_object)

    labeled_img = mask.copy()
    labeled_img = cv2.cvtColor(labeled_img, cv2.COLOR_GRAY2RGB)

    # Color each segment a different color, use a previously saved scale if available
    rand_color = color_palette(num=len(all_objects), saved=True)
    # Plot all segment contours
    for i, _ in enumerate(all_objects):
        cv2.drawContours(labeled_img, all_objects[i], -1, rand_color[i], params.line_thickness, lineType=8)
        # Store coordinates for labels
        label_coord_x.append(all_objects[i][0][0][0])
        label_coord_y.append(all_objects[i][0][0][1])

    # Label segments
    for i, _ in enumerate(all_objects):
        w = label_coord_x[i]
        h = label_coord_y[i]
        text = f"ID:{i}"
        cv2.putText(img=labeled_img, text=text, org=(w, h), fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                    fontScale=params.text_size, color=rand_color[i], thickness=2)

    _debug(visual=labeled_img, filename=os.path.join(params.debug_outdir, f"{params.device}_combined_segment_ids.png"))

    return labeled_img, all_objects



"""Fluorescence Analysis (NPQ parameter)."""
import os
import numpy as np
import pandas as pd
import xarray as xr
from math import ceil, floor
from plantcv.plantcv import params, outputs, fatal_error
from plantcv.plantcv._debug import _debug
from plantcv.plantcv.photosynthesis import reassign_frame_labels


def npq(ps_da_light, ps_da_dark, labeled_mask, n_labels=1, auto_fm=False, min_bin=0, max_bin="auto",
        measurement_labels=None, label=None):
    """
    Calculate and analyze non-photochemical quenching estimates from fluorescence image data.

    Inputs:
    ps_da_light        = Photosynthesis xarray DataArray that contains frame_label `Fmp` (ojip_light)
    ps_da_dark         = Photosynthesis xarray DataArray that contains frame_label `Fm` (ojip_dark)
    labeled_mask       = Labeled mask of objects (32-bit).
    n_labels           = Total number expected individual objects (default = 1).
    auto_fm            = Automatically calculate the frame with maximum fluorescence per label, otherwise
                         use a fixed frame for all labels (default = False).
    min_bin            = minimum bin value ("auto" or user input minimum value - must be an integer)
    max_bin            = maximum bin value ("auto" or user input maximum value - must be an integer)
    measurement_labels = labels for each measurement in ps_da_light, modifies the variable name of observations recorded
    label              = Optional label parameter, modifies the variable name of
                         observations recorded (default = pcv.params.sample_label).

    Returns:
    npq_global         = DataArray of NPQ values
    npq_chart          = Histograms of NPQ estimates

    :param ps_da_light: xarray.core.dataarray.DataArray
    :param ps_da_dark: xarray.core.dataarray.DataArray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param auto_fm: bool
    :param min_bin: int, str
    :param max_bin: int, str
    :param measurement_labels: list
    :param label: str
    :return npq_global: xarray.core.dataarray.DataArray
    :return npq_chart: altair.vegalite.v4.api.FacetChart
    """
    # Set labels
    labels = _set_labels(label, n_labels)

    if labeled_mask.shape != ps_da_light.shape[:2] or labeled_mask.shape != ps_da_dark.shape[:2]:
        fatal_error(f"Mask needs to have shape {ps_da_dark.shape[:2]}")

    if (measurement_labels is not None) and (len(measurement_labels) != ps_da_light.coords['measurement'].shape[0]):
        fatal_error('measurement_labels must be the same length as the number of measurements in `ps_da_light`')

    if ps_da_light.name.lower() != 'ojip_light' or ps_da_dark.name.lower() != 'ojip_dark':
        fatal_error(f"ps_da_light and ps_da_dark must be DataArrays with names 'ojip_light' and 'ojip_dark', "
                    f"respectively. Instead the inputs are: "
                    f"ps_da_light: {ps_da_light.name}, ps_da_dark: {ps_da_dark.name}"
                    )

    # Make an zeroed array of the same shape as the input DataArray
    npq_global = xr.zeros_like(ps_da_light, dtype=float)
    # Drop the frame_label coordinate
    npq_global = npq_global[:, :, 0, :].drop_vars('frame_label')

    # Make a copy of the labeled mask
    mask_copy = np.copy(labeled_mask)

    # If the labeled mask is a binary mask with values 0 and 255, convert to 0 and 1
    if len(np.unique(mask_copy)) == 2 and np.max(mask_copy) == 255:
        mask_copy = np.where(mask_copy == 255, 1, 0).astype(np.uint8)

    # Iterate over the label values 1 to n_labels
    for i in range(1, n_labels + 1):
        # Create a binary submask for each label
        submask = np.where(mask_copy == i, 255, 0).astype(np.uint8)

        # If auto_fm is True, reassign frame labels to choose the best Fm or Fm' for each labeled region
        if auto_fm:
            ps_da_light = reassign_frame_labels(ps_da=ps_da_light, mask=submask)
            ps_da_dark = reassign_frame_labels(ps_da=ps_da_dark, mask=submask)

        # Mask the Fm frame with the label submask
        fm = ps_da_dark.sel(measurement='t0', frame_label='Fm').where(submask > 0, other=0)
        # Calculate NPQ for the labeled region
        npq_lbl = ps_da_light.sel(frame_label='Fmp').groupby('measurement', squeeze=False).map(_calc_npq, fm=fm)

        # Drop the frame_label coordinate - not needed with xarray v2022.11.0+
        # npq_lbl = npq_lbl.drop_vars('frame_label')
        # Fill NaN values with 0 so that we can add DataArrays together
        npq_lbl = npq_lbl.fillna(0)
        # Add the NPQ values for this label to the NPQ DataArray
        npq_global = npq_global + npq_lbl

        # Record observations for each labeled region
        _add_observations(npq_da=npq_lbl, measurements=ps_da_light.measurement.values,
                          measurement_labels=measurement_labels, label=f"{labels[i - 1]}_{i}",
                          max_bin=max_bin, min_bin=min_bin)

    # Convert the labeled mask to a binary mask
    bin_mask = np.where(labeled_mask > 0, 255, 0)

    # Expand the binary mask to the same shape as the YII DataArray
    bin_mask = bin_mask[..., None]

    # Set the background values to NaN
    npq_global = npq_global.where(bin_mask > 0, other=np.nan)

    # drop coords identifying frames if they exist
    res = [i for i in list(npq_global.coords) if 'frame' in i]
    npq_global = npq_global.drop_vars(res)  # does not fail if res is []

    # Create a ridgeline plot of the NPQ values
    npq_chart = _ridgeline_plots(measurements=ps_da_light.measurement.values, measurement_labels=measurement_labels)

    # Plot/print dataarray
    _debug(visual=npq_global,
           filename=os.path.join(params.debug_outdir, str(params.device) + "_NPQ_dataarray.png"),
           col='measurement',
           col_wrap=int(np.ceil(npq_global.measurement.size / 4)),
           robust=True)

    # this only returns the last histogram..... xarray does not seem to support panels of histograms
    # but does support matplotlib subplots....
    return npq_global.squeeze(), npq_chart


def _set_labels(label, n_labels):
    """Create list of labels."""
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    # Set labels to label
    labels = label
    # If label is a string, make a list of labels
    if isinstance(label, str):
        labels = [label] * n_labels
    # If the length of the labels list is not equal to the number of labels, raise an error
    if len(labels) != n_labels:
        fatal_error(f"Number of labels ({len(labels)}) does not match number of objects ({n_labels})")
    return labels


def _calc_npq(fmp, fm):
    """NPQ = Fm/Fmp - 1."""
    out_flt = np.ones(shape=fm.shape) * np.nan
    fmp = np.squeeze(fmp)
    where_arr = np.logical_and(fm > 0, np.logical_and(fmp > 0, fm > fmp))
    div = np.divide(fm, fmp, out=out_flt, where=where_arr.to_numpy())
    sub = np.subtract(div, 1, out=out_flt.copy(), where=div.to_numpy() >= 1)
    return sub


def _create_histogram(npq_img, mlabel, min_bin, max_bin):
    """
    Compute histogram of NPQ.

    Inputs:
    npq_img     = numpy array of npq
    mlabel      = measurement label
    min_bin     = minimum bin value
    max_bin     = maximum bin value

    Returns:
    hist_df   = Dataframe of histogram
    npq_mode  = which non-zero bin has the maximum Fv/Fm value

    :param npq_img: numpy.ndarray
    :param mlabel: str
    :param obs: dict
    :param min_bin: int
    :param max_bin: int
    :return hist_df: pandas.DataFrame
    :return npq_mode: float
    """
    # Calculate the histogram of NPQ non-zero values
    npq_hist, npq_bins = np.histogram(npq_img[np.where(npq_img > 0)], 100, range=(min_bin, max_bin))
    # npq_bins is a bins + 1 length list of bin endpoints, so we need to calculate bin midpoints so that
    # the we have a one-to-one list of x (NPQ) and y (frequency) values.
    # To do this we add half the bin width to each lower bin edge x-value
    # midpoints = npq_bins[:-1] + 0.5 * np.diff(npq_bins)

    # Calculate which non-zero bin has the maximum Fv/Fm value
    npq_mode = npq_bins[np.argmax(npq_hist)]

    # Convert the histogram pixel counts to proportional frequencies
    npq_percent = (npq_hist / float(np.sum(npq_hist))) * 100

    # Create a dataframe
    hist_df = pd.DataFrame({'proportion of pixels (%)': npq_percent, mlabel: npq_bins[:-1]})

    return hist_df, npq_mode


def _add_observations(npq_da, measurements, measurement_labels, label, max_bin, min_bin):
    """Add observations for each labeled region."""
    # Auto calculate max_bin if set
    if isinstance(max_bin, str) and (max_bin.upper() == "AUTO"):
        max_bin = ceil(np.nanmax(npq_da))  # Auto bins will detect the max value to use for calculating labels/bins
    if isinstance(min_bin, str) and (min_bin.upper() == "AUTO"):
        min_bin = floor(np.nanmin(npq_da))  # Auto bins will detect the min value to use for calculating labels/bins

    # compute observations to store in Outputs, per labeled region
    npq_mean = npq_da.where(npq_da > 0).groupby('measurement').mean(['x', 'y']).values
    npq_median = npq_da.where(npq_da > 0).groupby('measurement').median(['x', 'y']).values
    npq_max = npq_da.where(npq_da > 0).groupby('measurement').max(['x', 'y']).values

    # Create variables to label traits based on measurement label in data array
    for i, mlabel in enumerate(measurements):
        if measurement_labels is not None:
            mlabel = measurement_labels[i]

        # mean value
        outputs.add_observation(sample=label, variable=f"npq_mean_{mlabel}", trait="mean npq value",
                                method='plantcv.plantcv.analyze.npq', scale='none', datatype=float,
                                value=float(npq_mean[i]), label='none')
        # median value
        outputs.add_observation(sample=label, variable=f"npq_median_{mlabel}", trait="median npq value",
                                method='plantcv.plantcv.analyze.npq', scale='none', datatype=float,
                                value=float(npq_median[i]), label='none')
        # max value
        outputs.add_observation(sample=label, variable=f"npq_max_{mlabel}", trait="peak npq value",
                                method='plantcv.plantcv.analyze.npq', scale='none', datatype=float,
                                value=float(npq_max[i]), label='none')

        hist_df, npq_mode = _create_histogram(npq_da.isel({'measurement': i}).values, mlabel, min_bin, max_bin)

        # mode value
        outputs.add_observation(sample=label, variable=f"npq_mode_{mlabel}", trait="mode npq value",
                                method='plantcv.plantcv.analyze.npq', scale='none', datatype=float,
                                value=float(npq_mode), label='none')
        # hist frequencies
        outputs.add_observation(sample=label, variable=f"npq_hist_{mlabel}", trait="frequencies",
                                method='plantcv.plantcv.analyze.npq', scale='none', datatype=list,
                                value=hist_df['proportion of pixels (%)'].values.tolist(),
                                label=np.around(hist_df[mlabel].values.tolist(), decimals=2).tolist())


def _ridgeline_plots(measurements, measurement_labels):
    """Create ridgeline plots of NPQ values."""
    npq_chart = None
    for i, mlabel in enumerate(measurements):
        if measurement_labels is not None:
            mlabel = measurement_labels[i]
        npq_chart = outputs.plot_dists(variable=f"npq_hist_{mlabel}")
        _debug(visual=npq_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_npq_hist.png'))
    return npq_chart



# Deprecation handling

import sys
from plantcv.plantcv import __version__ as version
from plantcv.plantcv import params


def deprecation_warning(warning):
    """Print out deprecation warning

    Inputs:
    warning = warning message text

    :param warning: str
    """
    warning_msg = f"DeprecationWarning: {warning} Current PlantCV version: {version}"
    if params.verbose is True:
        print(warning_msg, file=sys.stderr)



"""Analyzes the grayscale values of objects in an image."""
import os
import numpy as np
from plantcv.plantcv._debug import _debug
from plantcv.plantcv import params, outputs
from plantcv.plantcv.visualize import histogram
from plantcv.plantcv._helpers import _iterate_analysis


def grayscale(gray_img, labeled_mask, n_labels=1, bins=100, label=None):
    """Analyzes the grayscale values of a masked region of an image.

    Inputs:
    gray_img     = 8- or 16-bit grayscale image data.
    labeled_mask = Labeled mask of objects (32-bit).
    n_labels     = Total number expected individual objects (default = 1).
    bins         = Number of histogram bins.
    label        = Optional label parameter, modifies the variable name of
                   observations recorded (default = pcv.params.sample_label).

    Returns:
    analysis_image = Grayscale histogram image

    :param gray_img: numpy.ndarray
    :param labeled_mask: numpy.ndarray
    :param n_labels: int
    :param bins: int
    :param label: str
    :return analysis_image: altair.vegalite.v5.api.FacetChart
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    _ = _iterate_analysis(img=gray_img, labeled_mask=labeled_mask, n_labels=n_labels, label=label, function=_analyze_grayscale,
                          **{"bins": bins})
    gray_chart = outputs.plot_dists(variable="gray_frequencies")
    _debug(visual=gray_chart, filename=os.path.join(params.debug_outdir, str(params.device) + '_hue_hist.png'))
    return gray_chart


def _analyze_grayscale(img, mask, bins=100, label=None):
    """Analyzes the grayscale values of a masked region of an image.

    Inputs:
    img          = 8- or 16-bit grayscale image data.
    mask         = Labeled mask of objects (32-bit).
    bins         = Number of histogram bins.
    label        = optional label parameter, modifies the variable name of observations recorded (default = "default")

    Returns:
    img          = Input image

    :param gray_img: numpy.ndarray
    :param mask: numpy.ndarray
    :param bins: int
    :param label: str
    :return img: numpy.ndarray
    """
    # Save user debug setting
    debug = params.debug
    params.debug = None

    # Initialize output measurements
    hist_gray = [0] * bins
    bin_labels = list(range(0, bins))
    masked_gray_mean = 0
    masked_gray_median = 0
    masked_gray_std = 0

    # Skip empty masks
    if np.count_nonzero(mask) != 0:
        # calculate histogram
        if img.dtype == 'uint16':
            maxval = 65536
        else:
            maxval = 256

        masked_array = img[np.where(mask > 0)]
        masked_gray_mean = np.average(masked_array)
        masked_gray_median = np.median(masked_array)
        masked_gray_std = np.std(masked_array)

        # Calculate histogram
        _, hist_data = histogram(img, mask=mask, bins=bins, lower_bound=0, upper_bound=maxval, title=None,
                                 hist_data=True)

        bin_labels, hist_gray = hist_data["pixel intensity"].tolist(), hist_data['hist_count'].tolist()

        outputs.add_observation(sample=label, variable='gray_frequencies', trait='grayscale frequencies',
                                method='plantcv.plantcv.analyze.grayscale', scale='frequency', datatype=list,
                                value=hist_gray, label=bin_labels)
        outputs.add_observation(sample=label, variable='gray_mean', trait='grayscale mean',
                                method='plantcv.plantcv.analyze.grayscale', scale='none', datatype=float,
                                value=masked_gray_mean, label='none')
        outputs.add_observation(sample=label, variable='gray_median', trait='grayscale median',
                                method='plantcv.plantcv.analyze.grayscale', scale='none', datatype=float,
                                value=masked_gray_median, label='none')
        outputs.add_observation(sample=label, variable='gray_stdev', trait='grayscale standard deviation',
                                method='plantcv.plantcv.analyze.grayscale', scale='none', datatype=float,
                                value=masked_gray_std, label='none')
    # Restore user debug setting
    params.debug = debug

    return img



"""Analyzes the X and Y spatial distribution of objects in an image."""
import os
import numpy as np
from plantcv.plantcv import auto_crop, outputs, params
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _iterate_analysis


def distribution(labeled_mask, n_labels=1, direction="down", bin_size=100, hist_range="absolute", label=None):
    """Analyze the distribution of objects along an axis of an image.

    Parameters
    ----------
    labeled_mask : numpy.ndarray
        Labeled mask of objects (32-bit).
    n_labels : int, optional
        Total number expected individual objects, by default 1
    direction : str, optional
        Image axis to calculate the distribution of object pixels ("down" or "across"), by default "down"
    bin_size : int, optional
        Histogram bin size in pixels, by default 100
    hist_range : str, optional
        The histogram range can be set to the image dimensions ("absolute") or "relative" to each object, by default "absolute"
    label : str or None, optional
        Optional label parameter, modifies the variable name of observations recorded, by default pcv.params.sample_label

    Returns
    -------
    alt.vegalite.v5.api.FacetChart
        Facet chart of the object distribution histograms
    """
    # Increment the device counter
    params.device += 1

    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label

    # If direction is "across" then swap the x and y axes
    axis = "y"
    if direction == "across":
        axis = "x"
        labeled_mask = np.swapaxes(labeled_mask, 0, 1)

    # Create combined mask as "img" for iterative analysis input
    img = np.where(labeled_mask > 0, 255, 0).astype(np.uint8)

    # Iterate over each labeled object and analyze the distribution
    _ = _iterate_analysis(img=img, labeled_mask=labeled_mask, n_labels=n_labels, label=label,
                          function=_analyze_distribution,
                          **{"bin_size": bin_size, "direction": axis, "hist_range": hist_range})

    # Plot distributions
    dist_chart = outputs.plot_dists(variable=f"{axis}_frequencies")
    # Add plot labels
    dist_chart = dist_chart.properties(title=f"{axis}-axis distribution")

    # Display or save the debug plot
    _debug(visual=dist_chart, filename=os.path.join(params.debug_outdir, f"{params.device}_{axis}_distribution_hist.png"))

    return dist_chart


def _analyze_distribution(img, mask, direction="y", bin_size=100, hist_range="absolute", label=None):
    """Analyze the color properties of an image object
    Inputs:
    mask             = Binary mask made from selected contours
    bin_size         = Size in pixels of the histogram bins
    label            = optional label parameter, modifies the variable name of observations recorded

    Returns:
    distribution_image   = histogram output

    :param img: numpy.ndarray
    :param mask: numpy.ndarray
    :param bin_size: int
    :param label: str
    :return distribution_images: list
    """
    # Image not needed
    img -= 0

    # Store debug
    debug = params.debug
    params.debug = None

    # Autocrop the mask if hist_range is "relative" to set the scale to the object size
    if hist_range == "relative" and np.count_nonzero(mask) != 0:
        mask = auto_crop(img=mask, mask=mask, padding_x=0, padding_y=0, color="black")

    # Initialize output data
    # find the height, in pixels, for this image
    height = mask.shape[0]
    num_bins = height // bin_size

    # Initialize output measurements
    hist = np.zeros(num_bins)  # Histogram of the distribution
    counts = []  # Weighted list of bin values to calculate the median bin
    mean_bin = 0  # Mean bin value
    median_bin = 0  # Median bin value
    dist_std = 0  # Standard deviation of the distribution
    bin_labels = np.arange(num_bins) * bin_size  # Labels for the bins (pixel position)

    # Skip empty masks
    if np.count_nonzero(mask) != 0:
        # Calculate histogram
        for i in range(0, height, bin_size):
            # Extract a slice from the mask the width of bin_size at each step
            mask_slice = mask[i:min(i+bin_size, height), :]
            count = np.count_nonzero(mask_slice)  # Count white pixels
            bin_index = min(i // bin_size, num_bins - 1)  # Ensure index within range
            hist[bin_index] += count  # Add count to the bin
            counts += [bin_index * bin_size] * count

        # Calculate the median value distribution
        median_bin = np.median(counts)

        # Calculate the mean and standard deviation X and Y  value distribution
        mean_bin = np.mean(counts)
        dist_std = np.std(counts)

    # Save histograms
    outputs.add_observation(sample=label, variable=f'{direction}_frequencies', trait=f'{direction} frequencies',
                            method='plantcv.plantcv.analyze.distribution', scale='frequency', datatype=list,
                            value=hist.tolist(), label=bin_labels.tolist())

    # Save average measurements
    outputs.add_observation(sample=label, variable=f'{direction}_distribution_mean',
                            trait=f'{direction} distribution mean',
                            method='plantcv.plantcv.analyze.distribution', scale='pixels', datatype=float,
                            value=mean_bin, label='pixel')
    outputs.add_observation(sample=label, variable=f'{direction}_distribution_median',
                            trait=f'{direction} distribution median',
                            method='plantcv.plantcv.analyze.distribution', scale='pixel', datatype=float,
                            value=median_bin, label='pixel')
    outputs.add_observation(sample=label, variable=f'{direction}_distribution_std',
                            trait=f'{direction} distribution standard deviation',
                            method='plantcv.plantcv.analyze.distribution', scale='pixel', datatype=float,
                            value=dist_std, label='pixel')
    # Restore debug
    params.debug = debug

    return mask



# Fill in holes, flood fill

import numpy as np
import os
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rect_filter, _rect_replace
from plantcv.plantcv import fatal_error
from plantcv.plantcv import params
from skimage.segmentation import flood_fill


def floodfill(bin_img, points, value=0, roi=None):
    """
    Flood fills holes in a binary mask

    Inputs:
    bin_img = Binary image data
    points  = list of seed points to start flood fill e.g. ([(x,y)])
    value   = value from 0-255 (default value=0)
    roi     = optional rectangular ROI (note points are relative to this ROI)

    Returns:
    filled_img = image with objects filled

    :param bin_img: numpy.ndarray
    :param points: list
    :param value: float
    :param roi: plantcv.plantcv.Objects
    :return filled_img: numpy.ndarray
    """
    # Make sure the image is binary
    if len(np.shape(bin_img)) != 2 or len(np.unique(bin_img)) != 2:
        fatal_error("Image is not binary")

    # Cast binary image to boolean
    bool_img = bin_img.astype(bool)

    # Flood fill holes
    points1 = list(map(lambda sub: (sub[1], sub[0]), points))

    for x in points1:
        sub_bool_img = _rect_filter(bool_img, roi=roi, function=flood_fill,
                                    **{"seed_point": x, "new_value": value})
        bool_img = _rect_replace(bool_img, sub_bool_img, roi)

    # Cast boolean image to binary and make a copy of the binary image for returning
    filled_img = np.copy(bool_img.astype(np.uint8) * 255)

    _debug(visual=filled_img,
           filename=os.path.join(params.debug_outdir, str(params.device) + '_flood_fill' + '.png'),
           cmap='gray')

    return filled_img



# Read dataset of images

import os
import re


def read_dataset(source_path, pattern='', sort=True):
    """
    Read a dataset of images as a list of paths.
    Inputs:
    source_path  = Path to the directory containing the images
    pattern      = Optional, return only filenames containing the pattern
    sort         = True by default, sorts the paths alphabetically
    Returns:
    dataset = List of paths to the images in the source path
    :param source_path: str
    :param pattern: str
    :param sort: bool
    :return dataset: list
    """
    if not os.path.exists(source_path):
        raise IOError(f"Directory does not exist: {source_path}")

    img_path_list = []
    img_extensions = ['.png', '.jpg', '.jpeg', '.tif', '.tiff', '.gif']

    for root, _, files in os.walk(source_path):
        for file in files:
            # Look for images that contain [pattern] in the name
            if re.search(pattern, file):
                # Check file type so that only images get selected
                _, ext = os.path.splitext(file)
                if ext.lower() in img_extensions:
                    img_path_list.append(os.path.join(root, file))

    if sort is True:
        img_path_list = sorted(img_path_list)

    return img_path_list



"""Automatically detect color cards.

Algorithm written by mtwatso2-eng (github). Updated and implemented into PlantCV by Haley Schuhl.
"""
import os
import cv2
import math
import numpy as np
from plantcv.plantcv import params, outputs, fatal_error, deprecation_warning
from plantcv.plantcv._debug import _debug
from plantcv.plantcv._helpers import _rgb2gray, _cv2_findcontours, _object_composition, _rect_filter, _rect_replace


def _is_square(contour, min_size, aspect_ratio=1.27, solidity=.8):
    """Return list of contours, based on aspect ratio and solidity.

    Parameters
    ----------
    contour : list
        OpenCV contour.
    min_size : int
        Minimum object size to be considered
    aspect_ratio : float
        Filter contours below a given aspect ratio
    solidity : float
        Filter contours below a given solidity

    Returns
    -------
    bool
        True if the contour is square, False otherwise.
    """
    # Take reciprocal if aspect_ratio is smaller than 1
    aspect_ratio = max([aspect_ratio, 1]) / min([aspect_ratio, 1])

    return (cv2.contourArea(contour) > min_size and
            # Test that the Aspect Ratio (default 1.27)
            # ratio between the width and height of minAreaRect
            # which is like a bounding box but will consider rotation
            max(cv2.minAreaRect(contour)[1]) / min(cv2.minAreaRect(contour)[1]) < aspect_ratio and
            # Test that the Solidity (default 0.8)
            # Compare minAreaRect area to the actual contour area, a chip should be mostly solid
            (cv2.contourArea(contour) / np.prod(cv2.minAreaRect(contour)[1])) > solidity)


def _get_contour_sizes(contours):
    """Get the shape and size of all contours.

    Parameters
    ----------
    contours : list
        List of OpenCV contours.

    Returns
    -------
    list
        Contour areas, widths, and heights.
    """
    # Initialize chip shape lists
    marea, mwidth, mheight = [], [], []
    # Loop over our contours and size data about them
    for cnt in contours:
        marea.append(cv2.contourArea(cnt))
        _, wh, _ = cv2.minAreaRect(cnt)  # Rotated rectangle
        mwidth.append(wh[0])
        mheight.append(wh[1])
    return marea, mwidth, mheight


def _draw_color_chips(rgb_img, new_centers, radius):
    """Create labeled mask and debug image of color chips.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        Input RGB image data containing a color card.
    new_centers : numpy.array
        Chip centers after transformation.
    radius : int
        Radius of circles to draw on the color chips.

    Returns
    -------
    list
        Labeled mask and debug image.
    """
    # Create blank img for drawing the labeled color card mask
    labeled_mask = np.zeros(rgb_img.shape[0:2])
    debug_img = np.copy(rgb_img)

    # Loop over the new chip centers and draw them on the RGB image and labeled mask
    for i, pt in enumerate(new_centers):
        cv2.circle(labeled_mask, new_centers[i], radius, (i + 1) * 10, -1)
        cv2.circle(debug_img, new_centers[i], radius, (255, 255, 0), -1)
        cv2.putText(debug_img, text=str(i), org=pt, fontScale=params.text_size, color=(0, 0, 0),
                    fontFace=cv2.FONT_HERSHEY_SIMPLEX, thickness=params.text_thickness)
    return labeled_mask, debug_img


def _check_point_per_chip(contours, centers, debug_img):
    """Check that each detected chip ends up with exactly 1 mask inside it

    Parameters
    ----------
    contours : list
             OpenCV contour
    centers : numpy.ndarray
             (X, Y) points of the center of each prospective mask to make
    debug_img : numpy.ndarray
             Debug image to show the problem if a fatal error is generated

    Returns
    -------
    No returns

    Raises
    ------
    fatal_error
          If any contour does not have exactly 1 mask center inside it (not on edge)
    """
    contour_has_n_points = []
    for cont in contours:
        bools = []
        for pt in centers:
            # -1 is outside, 0 is on line, 1 is inside
            bools.append(cv2.pointPolygonTest(cont, (int(pt[0]), int(pt[1])), False) == 1)
        contour_has_n_points.append(sum(bools))

    if any(n != 1 for n in contour_has_n_points):
        _debug(visual=debug_img, filename=os.path.join(params.debug_outdir, f'{params.device}_color_card.png'))
        fatal_error("Centers do not map 1 to 1 with detected color chips")


def _check_corners(img, corners):
    """Check that corners are within an image
    Parameters
    ----------
    img: numpy.ndarray
         An image
    corners: numpy.ndarray
         Corners for some object
    """
    dim = np.shape(img)
    for pt in corners:
        if pt[0] > dim[1] or pt[1] > dim[0] or pt[0] < 0 or pt[1] < 0:
            fatal_error('Color card corners could not be detected accurately')


def _color_card_detection(rgb_img, **kwargs):
    """Algorithm to automatically detect a color card.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        Input RGB image data containing a color card.
    label : str, optional
        modifies the variable name of observations recorded (default = pcv.params.sample_label).
    **kwargs
        Other keyword arguments passed to cv2.adaptiveThreshold and cv2.circle.

        Valid keyword arguments:
        adaptive_method: 0 (mean) or 1 (Gaussian) (default = 1)
        block_size: int (default = 51)
        radius: int (default = 20)
        min_size: int (default = 1000)
        aspect_ratio: countour squareness filters (default 1.27)
        solidity: contour squareness filters (default 0.8)

    Returns
    -------
    list
        Labeled mask of chips, debug img, detected chip areas, chip heights, chip widths, bounding box mask
    """
    # Get keyword arguments and set defaults if not set
    min_size = kwargs.get("min_size", 1000)  # Minimum size for _is_square chip filtering
    radius = kwargs.get("radius", 20)  # Radius of circles to draw on the color chips
    adaptive_method = kwargs.get("adaptive_method", 1)  # cv2.adaptiveThreshold method
    block_size = kwargs.get("block_size", 51)  # cv2.adaptiveThreshold block size
    aspect_ratio = kwargs.get("aspect_ratio", 1.27)  # _is_square aspect-ratio filtering
    solidity = kwargs.get("solidity", 0.8)  # _is_square solidity filtering

    # Throw a fatal error if block_size is not odd or greater than 1
    if not (block_size % 2 == 1 and block_size > 1):
        fatal_error('block_size parameter must be an odd int greater than 1.')

    # Hard code since we don't currently support other color cards
    nrows = 6
    ncols = 4

    # Convert to grayscale, threshold, and findContours
    imgray = _rgb2gray(rgb_img=rgb_img)
    gaussian = cv2.GaussianBlur(imgray, (11, 11), 0)
    thresh = cv2.adaptiveThreshold(gaussian, 255, adaptive_method,
                                   cv2.THRESH_BINARY_INV, block_size, 2)
    contours, _ = cv2.findContours(thresh, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)

    # Filter contours, keep only square-shaped ones
    filtered_contours = [contour for contour in contours if _is_square(contour, min_size, aspect_ratio, solidity)]
    # Calculate median area of square contours
    target_square_area = np.median([cv2.contourArea(cnt) for cnt in filtered_contours])
    # Filter contours again, keep only those within 20% of median area
    filtered_contours = [contour for contour in filtered_contours if
                         (0.8 < (cv2.contourArea(contour) / target_square_area) < 1.2)]

    # Throw a fatal error if no color card found
    if len(filtered_contours) == 0:
        fatal_error('No color card found')

    # Draw filtered contours on debug img
    debug_img = np.copy(rgb_img)
    cv2.drawContours(debug_img, filtered_contours, -1, color=(255, 50, 250), thickness=params.line_thickness)
    # Find the bounding box of the detected chips
    x, y, w, h = cv2.boundingRect(np.vstack(filtered_contours))

    # Draw the bound box rectangle
    boundind_mask = cv2.rectangle(np.zeros(rgb_img.shape[0:2]), (x, y), (x + w, y + h), (255), -1).astype(np.uint8)

    # Initialize chip shape lists
    marea, mwidth, mheight = _get_contour_sizes(filtered_contours)

    # Concatenate all contours into one array and find the minimum area rectangle
    rect = np.concatenate([[np.array(cv2.minAreaRect(i)[0]).astype(int)] for i in filtered_contours])
    rect = cv2.minAreaRect(rect)
    # Get the corners of the rectangle
    corners = np.array(np.intp(cv2.boxPoints(rect)))
    # Check that corners are in image
    _check_corners(imgray, corners)
    # Determine which corner most likely contains the white chip
    white_index = np.argmin([np.mean(math.dist(rgb_img[corner[1], corner[0], :], (255, 255, 255))) for corner in corners])
    corners = corners[np.argsort([math.dist(corner, corners[white_index]) for corner in corners])[[0, 1, 3, 2]]]
    # Increment amount is arbitrary, cell distances rescaled during perspective transform
    increment = 100
    centers = [[int(0 + i * increment), int(0 + j * increment)] for j in range(nrows) for i in range(ncols)]

    # Find the minimum area rectangle of the chip centers
    new_rect = cv2.minAreaRect(np.array(centers))
    # Get the corners of the rectangle
    box_points = cv2.boxPoints(new_rect).astype("float32")
    # Calculate the perspective transform matrix from the minimum area rectangle
    m_transform = cv2.getPerspectiveTransform(box_points, corners.astype("float32"))
    # Transform the chip centers using the perspective transform matrix
    new_centers = cv2.transform(np.array([centers]), m_transform)[0][:, 0:2]
    # Create labeled mask and debug image of color chips
    labeled_mask, debug_img = _draw_color_chips(debug_img, new_centers, radius)
    # Check that new centers are inside each unique filtered_contour
    _check_point_per_chip(filtered_contours, new_centers, debug_img)

    return labeled_mask, debug_img, marea, mheight, mwidth, boundind_mask


def _set_size_scale_from_chip(color_chip_width, color_chip_height, color_chip_size):
    """Set the size scaling factors in Params from the known size of a given color card target.

    Parameters
    ----------
    color_chip_width : float
        Width in pixels of the detected color chips
    color_chip_height : float
        Height in pixels of the detected color chips
    color_chip_size : str, tuple
        Type of supported color card target ("classic", "passport", "nano", "mini", or "cameratrax"),
        or a tuple of (width, height) of the color card chip real-world dimensions in milimeters.
    """
    # Define known color chip dimensions, all in milimeters
    card_types = {
        "CLASSIC": {
            "chip_width": 40,
            "chip_height": 40
        },
        "PASSPORT": {
            "chip_width": 12,
            "chip_height": 12
        },
        "CAMERATRAX": {
            "chip_width": 11,
            "chip_height": 11
        },
        "NANO": {
            "chip_width": 4,
            "chip_height": 3
        },
        "MINI": {
            "chip_width": 12,
            "chip_height": 12
        }
    }

    # Check if user provided a valid color card type
    if type(color_chip_size) is str and color_chip_size.upper() in card_types:
        # Set size scaling parameters, match larger dimensions with each other
        obs_chip_dims = [color_chip_width, color_chip_height]
        known_chip_dims = [card_types[color_chip_size.upper()]["chip_width"],
                           card_types[color_chip_size.upper()]["chip_height"]]
        params.px_width = max(known_chip_dims) / max(obs_chip_dims)
        params.px_height = min(known_chip_dims) / min(obs_chip_dims)
    # If not, check to make sure custom dimensions provided are numeric
    else:
        try:
            # Set size scaling parameters
            params.px_width = float(color_chip_size[0]) / color_chip_width
            params.px_height = float(color_chip_size[1]) / color_chip_height
        # Fail if provided color_chip_size is not supported
        except ValueError:
            fatal_error(f"Invalid input '{color_chip_size}'. Choose from {list(card_types.keys())}\
            or provide your color card chip dimensions explicitly")
        # Fail if provided color_chip_size is integer rather than tuple
        except TypeError:
            fatal_error(f"Invalid input '{color_chip_size}'. Choose from {list(card_types.keys())}\
            or provide your color card chip dimensions explicitly as a tuple e.g. color_chip_size=(10,10).")
    # If size scaling successful, set units to millimeters
    params.unit = "mm"


def mask_color_card(rgb_img, **kwargs):
    """Automatically detect a color card and create bounding box mask of the chips detected.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        Input RGB image data containing a color card.
    **kwargs
        Other keyword arguments passed to cv2.adaptiveThreshold and cv2.circle.

        Valid keyword arguments:
        adaptive_method: 0 (mean) or 1 (Gaussian) (default = 1)
        block_size: int (default = 51)
        radius: int (default = 20)
        min_size: int (default = 1000)

    Returns
    -------

    numpy.ndarray
        Binary bounding box mask of the detected color card chips
    """
    _, _, _, _, _, bounding_mask = _color_card_detection(rgb_img, **kwargs)

    if params.debug is not None:
        # Find contours
        cnt, cnt_str = _cv2_findcontours(bin_img=bounding_mask)

        # Consolidate contours
        obj = _object_composition(contours=cnt, hierarchy=cnt_str)
        bb_debug = cv2.drawContours(np.copy(rgb_img), [obj], -1, (255, 0, 255), params.line_thickness)

        # Debug image handling
        _debug(visual=bb_debug, filename=os.path.join(params.debug_outdir, f'{params.device}_color_card.png'))

    return bounding_mask


def detect_color_card(rgb_img, label=None, color_chip_size=None, roi=None, **kwargs):
    """Automatically detect a Macbeth ColorChecker style color card.

    Parameters
    ----------
    rgb_img : numpy.ndarray
        Input RGB image data containing a color card.
    label : str, optional
        modifies the variable name of observations recorded (default = pcv.params.sample_label).
    color_chip_size: str, tuple, optional
        "passport", "classic", "nano", "mini", or "cameratrax"; or tuple formatted (width, height)
        in millimeters (default = None)
    roi : plantcv.plantcv.Objects, optional
        A rectangular ROI as returned from pcv.roi.rectangle to detect a color card only in that region.
    **kwargs
        Other keyword arguments passed to cv2.adaptiveThreshold and cv2.circle.

        Valid keyword arguments:
        adaptive_method: 0 (mean) or 1 (Gaussian) (default = 1)
        block_size: int (default = 51)
        radius: int (default = 20)
        min_size: int (default = 1000)
        aspect_ratio: float (default = 1.27)
        solidity: float (default = 0.8)


    Returns
    -------
    numpy.ndarray
        Labeled mask of chips.
    """
    # Set lable to params.sample_label if None
    if label is None:
        label = params.sample_label
    deprecation_warning(
        "The 'label' parameter is no longer utilized, since color chip size is now metadata. "
        "It will be removed in PlantCV v5.0."
        )
    # apply _color_card_detection within bounding box
    sub_mask, debug_img, marea, mheight, mwidth, _ = _rect_filter(rgb_img,
                                                                  roi,
                                                                  function=_color_card_detection,
                                                                  **kwargs)
    # slice sub_mask from bounding box into mask of original image size
    empty_mask = np.zeros((np.shape(rgb_img)[0], np.shape(rgb_img)[1]))
    labeled_mask = _rect_replace(empty_mask, sub_mask, roi)

    # Create dataframe for easy summary stats
    chip_size = np.median(marea)
    chip_height = np.median(mheight)
    chip_width = np.median(mwidth)

    # Save out chip size for pixel to mm standardization
    outputs.add_metadata(term="median_color_chip_size", datatype=float, value=chip_size)
    outputs.add_metadata(term="median_color_chip_width", datatype=float, value=chip_width)
    outputs.add_metadata(term="median_color_chip_height", datatype=float, value=chip_height)

    # Set size scaling factor if card type is provided
    if color_chip_size:
        _set_size_scale_from_chip(color_chip_height=chip_height, color_chip_width=chip_width, color_chip_size=color_chip_size)

    # Debugging
    _debug(visual=debug_img, filename=os.path.join(params.debug_outdir, f'{params.device}_color_card.png'))

    return labeled_mask



# Visualize a scatter plot of pixels

import numpy as np
import cv2
from matplotlib import pyplot as plt
from plantcv.plantcv import fatal_error, params
from plantcv.plantcv.readimage import readimage
from plantcv.plantcv._helpers import _rgb2lab, _rgb2hsv, _rgb2cmyk, _rgb2gray


MAX_MARKER_SIZE = 20
IMG_WIDTH = 128


# functions to get a given channel with parameters compatible
# with rgb2gray_lab, rgb2gray_hsv, and rgb2gray_cmyk to use in the dict
def _get_R(rgb_img, _):
    """Get the red channel from a RGB image."""
    return rgb_img[:, :, 2]


def _get_G(rgb_img, _):
    """Get the green channel from a RGB image."""
    return rgb_img[:, :, 1]


def _get_B(rgb_img, _):
    """Get the blue channel from a RGB image."""
    return rgb_img[:, :, 0]


def _get_gray(rgb_img, _):
    """Get the gray scale transformation of a RGB image."""
    return _rgb2gray(rgb_img=rgb_img)


def _get_index(rgb_img, _):
    """Get a vector with linear indices of the pixels in an image."""
    h, w, _ = rgb_img.shape
    return np.arange(h*w)


def _not_valid(*args):
    """Error for a non valid channel."""
    return fatal_error("channel not valid, use R, G, B, l, a, b, h, s, v, c, m, y, k, gray, or index")


def pixel_scatter_plot(paths_to_imgs, x_channel, y_channel):
    """
    Plot a 2D pixel scatter plot visualization for a dataset of images.
    The horizontal and vertical coordinates are defined by the intensity of the
    pixels in the specified channels.
    The color of each dot is given by the original RGB color of the pixel.

    Inputs:
    paths_to_imgs  = List of paths to the images
    x_channel      = Channel to use for the horizontal coordinate of the scatter plot.
                     Options:  'R', 'G', 'B', 'l', 'a', 'b', 'h', 's', 'v', 'c', 'm', 'y', 'k', 'gray', and 'index'
    y_channel      = Channel to use for the vertical coordinate of the scatter plot.
                     Options:  'R', 'G', 'B', 'l', 'a', 'b', 'h', 's', 'v', 'c', 'm', 'y', 'k', 'gray', and 'index'

    Returns:
    fig = matplotlib pyplot Figure object of the visualization
    ax  = matplotlib pyplot Axes object of the visualization

    :param paths_to_imgs: list of str
    :param x_channel: str
    :param y_channel: str
    :return fig: matplotlib.pyplot Figure object
    :return ax: matplotlib.pyplot Axes object
    """
    # dictionary returns the function that gets the required image channel
    channel_dict = {
        'R': _get_R,
        'G': _get_G,
        'B': _get_B,
        'l': _rgb2lab,
        'a': _rgb2lab,
        'b': _rgb2lab,
        'gray': _get_gray,
        'h': _rgb2hsv,
        's': _rgb2hsv,
        'v': _rgb2hsv,
        'index': _get_index,
        'c': _rgb2cmyk,
        'm': _rgb2cmyk,
        'y': _rgb2cmyk,
        'k': _rgb2cmyk
    }

    # store debug mode
    debug = params.debug
    params.debug = None

    N = len(paths_to_imgs)

    fig, ax = plt.subplots()
    # load and plot the set of images sequentially
    for p in paths_to_imgs:
        img, _, _ = readimage(filename=p)
        h, _, c = img.shape

        # resizing to predetermined width to reduce the number of pixels
        ratio = h/IMG_WIDTH
        img_height = int(IMG_WIDTH*ratio)
        # nearest interpolation avoids mixing pixel values
        sub_img = cv2.resize(img, (IMG_WIDTH, img_height), interpolation=cv2.INTER_NEAREST)

        # organize the channels as RGB to use as facecolor for the markers
        sub_img_rgb = cv2.cvtColor(sub_img, cv2.COLOR_BGR2RGB)
        fcolors = sub_img_rgb.reshape(img_height*IMG_WIDTH, c)/255

        # get channels
        sub_img_x_ch = channel_dict.get(x_channel, _not_valid)(sub_img, x_channel)
        sub_img_y_ch = channel_dict.get(y_channel, _not_valid)(sub_img, y_channel)

        ax.scatter(sub_img_x_ch.reshape(-1),
                   sub_img_y_ch.reshape(-1),
                   alpha=0.05, s=MAX_MARKER_SIZE/N,
                   edgecolors=None, facecolors=fcolors)

    plt.xlabel(x_channel)
    plt.ylabel(y_channel)

    # reset debug
    params.debug = debug

    return fig, ax
